#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime
from unittest import mock

import pytest

from airflow.exceptions import AirflowException, AirflowProviderDeprecationWarning
from airflow.providers.common.compat.openlineage.facet import Dataset
from airflow.providers.google.cloud.transfers.gcs_to_gcs import WILDCARD, GCSToGCSOperator

TASK_ID = "test-gcs-to-gcs-operator"
TEST_BUCKET = "test-bucket"
PREFIX = "TEST"
SOURCE_FOLDER = "test_object"
SOURCE_OBJECTS_NO_FILE = [""]
SOURCE_OBJECTS_TWO_EMPTY_STRING = ["", ""]
SOURCE_OBJECTS_SINGLE_FILE = [f"{SOURCE_FOLDER}/file1.txt"]
SOURCE_OBJECTS_MULTIPLE_FILES = [f"{SOURCE_FOLDER}/file1.txt", f"{SOURCE_FOLDER}/file2.txt"]
SOURCE_OBJECTS_LIST = [
    f"{SOURCE_FOLDER}/file1.txt",
    f"{SOURCE_FOLDER}/file2.txt",
    f"{SOURCE_FOLDER}/file3.json",
]

SOURCE_OBJECT_WILDCARD_PREFIX = "*test_object"
SOURCE_OBJECT_WILDCARD_SUFFIX = "test_object*"
SOURCE_OBJECT_WILDCARD_MIDDLE = "test*object"
SOURCE_OBJECT_WILDCARD_FILENAME = "test_object*.txt"
SOURCE_OBJECT_WILDCARD_NESTED = "test_object/file1.txt"
SOURCE_OBJECT_NO_WILDCARD = "test_object.txt"
SOURCE_OBJECT_MULTIPLE_WILDCARDS = "csv/*/test_*.csv"
DESTINATION_BUCKET = "archive"
DESTINATION_OBJECT = "foo/bar/"
DESTINATION_OBJECT_PREFIX = "foo/bar"
DELIMITER = ".json"

MOD_TIME_1 = datetime(2016, 1, 1)
MOD_TIME_2 = datetime(2019, 1, 1)


# TODO: After deprecating delimiter and wildcards in source objects,
#       implement reverted changes from the first commit of PR #31261
class TestGoogleCloudStorageToCloudStorageOperator:
    """
    Tests the three use-cases for the wildcard operator. These are
    no_prefix: *test_object
    no_suffix: test_object*
    prefix_and_suffix: test*object
    Also tests the destination_object as prefix when the wildcard is used.
    """

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_no_prefix(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_PREFIX,
            destination_bucket=DESTINATION_BUCKET,
        )

        operator.execute(None)
        mock_hook.return_value.list.assert_called_once_with(TEST_BUCKET, prefix="", delimiter="test_object")

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_no_suffix(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_SUFFIX,
            destination_bucket=DESTINATION_BUCKET,
        )

        operator.execute(None)
        mock_hook.return_value.list.assert_called_once_with(TEST_BUCKET, prefix="test_object", delimiter="")

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_wildcard_with_replace_flag_false(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_SUFFIX,
            destination_bucket=DESTINATION_BUCKET,
            replace=False,
        )

        operator.execute(None)
        mock_calls = [
            mock.call(TEST_BUCKET, prefix="test_object", delimiter=""),
            mock.call(DESTINATION_BUCKET, prefix="test_object", delimiter="", match_glob=None),
        ]
        mock_hook.return_value.list.assert_has_calls(mock_calls)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_no_wildcard_with_replace_flag_false(self, mock_hook):
        mock_hook.return_value.list.side_effect = [[], [SOURCE_OBJECT_NO_WILDCARD]]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=DESTINATION_BUCKET,
            replace=False,
        )

        operator.execute(None)
        mock_calls = [
            mock.call(TEST_BUCKET, prefix=SOURCE_OBJECT_NO_WILDCARD, delimiter=None, match_glob=None),
            mock.call(DESTINATION_BUCKET, prefix=SOURCE_OBJECT_NO_WILDCARD, delimiter=None, match_glob=None),
        ]
        mock_hook.return_value.list.assert_has_calls(mock_calls)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_copy_file_with_exact_match(self, mock_hook):
        SOURCE_FILES = [
            "test_object.txt",
            "test_object.txt.abc",
            "test_object.txt.copy/",
            "test_object.txt.folder/",
        ]
        mock_hook.return_value.list.return_value = SOURCE_FILES
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=DESTINATION_BUCKET,
            exact_match=True,
        )

        operator.execute(None)
        mock_calls = [
            mock.call(TEST_BUCKET, prefix="test_object.txt", delimiter=None, match_glob=None),
        ]
        mock_hook.return_value.list.assert_has_calls(mock_calls)
        mock_hook.return_value.rewrite.assert_has_calls(
            [
                mock.call(TEST_BUCKET, "test_object.txt", DESTINATION_BUCKET, "test_object.txt"),
            ]
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_copy_file_with_exact_match_destination(self, mock_hook):
        SOURCE_FILES = [
            "test_object.txt",
            "test_object.txt.abc",
            "test_object.txt.copy/",
            "test_object.txt.folder/",
        ]
        DESTINATION_OBJ = f"{DESTINATION_OBJECT_PREFIX}/test_object.txt"

        mock_hook.return_value.list.return_value = SOURCE_FILES
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=DESTINATION_OBJ,
            exact_match=True,
        )

        operator.execute(None)
        mock_calls = [
            mock.call(TEST_BUCKET, prefix="test_object.txt", delimiter=None, match_glob=None),
        ]
        mock_hook.return_value.list.assert_has_calls(mock_calls)

        mock_calls_rewrite = [
            mock.call(TEST_BUCKET, "test_object.txt", DESTINATION_BUCKET, DESTINATION_OBJ),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls_rewrite)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_prefix_and_suffix(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_MIDDLE,
            destination_bucket=DESTINATION_BUCKET,
        )

        operator.execute(None)
        mock_hook.return_value.list.assert_called_once_with(TEST_BUCKET, prefix="test", delimiter="object")

    # copy with wildcard
    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_wildcard_with_destination_object(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST[:-1]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_FILENAME,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=DESTINATION_OBJECT_PREFIX,
        )

        operator.execute(None)
        mock_calls = [
            mock.call(TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "foo/bar/file1.txt"),
            mock.call(TEST_BUCKET, "test_object/file2.txt", DESTINATION_BUCKET, "foo/bar/file2.txt"),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_wildcard_with_destination_object_retained_prefix(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST[:-1]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_FILENAME,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=f"{DESTINATION_OBJECT_PREFIX}/{SOURCE_OBJECT_WILDCARD_SUFFIX[:-1]}",
        )

        operator.execute(None)
        mock_calls_retained = [
            mock.call(
                TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "foo/bar/test_object/file1.txt"
            ),
            mock.call(
                TEST_BUCKET, "test_object/file2.txt", DESTINATION_BUCKET, "foo/bar/test_object/file2.txt"
            ),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls_retained)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_wildcard_without_destination_object(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST[:-1]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_FILENAME,
            destination_bucket=DESTINATION_BUCKET,
        )

        operator.execute(None)
        mock_calls_none = [
            mock.call(TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "test_object/file1.txt"),
            mock.call(TEST_BUCKET, "test_object/file2.txt", DESTINATION_BUCKET, "test_object/file2.txt"),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls_none)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_wildcard_empty_destination_object(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST[:-1]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_FILENAME,
            destination_bucket=DESTINATION_BUCKET,
            destination_object="",
        )

        operator.execute(None)
        mock_calls_empty = [
            mock.call(TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "/file1.txt"),
            mock.call(TEST_BUCKET, "test_object/file2.txt", DESTINATION_BUCKET, "/file2.txt"),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls_empty)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_wildcard_with_destination_object_rename(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_SINGLE_FILE
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_NESTED,
            destination_bucket=DESTINATION_BUCKET,
            destination_object="test_file.txt",
        )

        operator.execute(None)
        mock_calls = [
            mock.call(TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "test_file.txt"),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_last_modified_time(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST[:-1]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_FILENAME,
            destination_bucket=DESTINATION_BUCKET,
            last_modified_time=None,
        )

        operator.execute(None)
        mock_calls_none = [
            mock.call(TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "test_object/file1.txt"),
            mock.call(TEST_BUCKET, "test_object/file2.txt", DESTINATION_BUCKET, "test_object/file2.txt"),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls_none)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_wc_with_last_modified_time_with_all_true_cond(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST[:-1]
        mock_hook.return_value.is_updated_after.side_effect = [True, True, True]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_FILENAME,
            destination_bucket=DESTINATION_BUCKET,
            last_modified_time=MOD_TIME_1,
        )

        operator.execute(None)
        mock_calls_none = [
            mock.call(TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "test_object/file1.txt"),
            mock.call(TEST_BUCKET, "test_object/file2.txt", DESTINATION_BUCKET, "test_object/file2.txt"),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls_none)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_wc_with_last_modified_time_with_one_true_cond(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST[:-1]
        mock_hook.return_value.is_updated_after.side_effect = [True, False, False]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_FILENAME,
            destination_bucket=DESTINATION_BUCKET,
            last_modified_time=MOD_TIME_1,
        )

        operator.execute(None)
        mock_hook.return_value.rewrite.assert_called_once_with(
            TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "test_object/file1.txt"
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_wc_with_no_last_modified_time(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST[:-1]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_FILENAME,
            destination_bucket=DESTINATION_BUCKET,
            last_modified_time=None,
        )

        operator.execute(None)
        mock_calls_none = [
            mock.call(TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "test_object/file1.txt"),
            mock.call(TEST_BUCKET, "test_object/file2.txt", DESTINATION_BUCKET, "test_object/file2.txt"),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls_none)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_no_prefix_with_last_modified_time_with_true_cond(self, mock_hook):
        mock_hook.return_value.is_updated_after.return_value = True
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=SOURCE_OBJECT_NO_WILDCARD,
            last_modified_time=MOD_TIME_1,
        )

        operator.execute(None)
        mock_hook.return_value.rewrite.assert_called_once_with(
            TEST_BUCKET, SOURCE_OBJECT_NO_WILDCARD, DESTINATION_BUCKET, SOURCE_OBJECT_NO_WILDCARD
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_no_prefix_with_maximum_modified_time_with_true_cond(self, mock_hook):
        mock_hook.return_value.is_updated_before.return_value = True
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=SOURCE_OBJECT_NO_WILDCARD,
            maximum_modified_time=MOD_TIME_1,
        )

        operator.execute(None)
        mock_hook.return_value.rewrite.assert_called_once_with(
            TEST_BUCKET, SOURCE_OBJECT_NO_WILDCARD, DESTINATION_BUCKET, SOURCE_OBJECT_NO_WILDCARD
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_exe_last_modified_time_and_maximum_modified_time_with_true_cond(self, mock_hook):
        mock_hook.return_value.is_updated_between.return_value = True
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=SOURCE_OBJECT_NO_WILDCARD,
            last_modified_time=MOD_TIME_1,
            maximum_modified_time=MOD_TIME_2,
        )

        operator.execute(None)
        mock_hook.return_value.rewrite.assert_called_once_with(
            TEST_BUCKET, SOURCE_OBJECT_NO_WILDCARD, DESTINATION_BUCKET, SOURCE_OBJECT_NO_WILDCARD
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_no_prefix_with_no_last_modified_time(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=SOURCE_OBJECT_NO_WILDCARD,
            last_modified_time=None,
        )

        operator.execute(None)
        mock_hook.return_value.rewrite.assert_called_once_with(
            TEST_BUCKET, SOURCE_OBJECT_NO_WILDCARD, DESTINATION_BUCKET, SOURCE_OBJECT_NO_WILDCARD
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_no_prefix_with_last_modified_time_with_false_cond(self, mock_hook):
        mock_hook.return_value.is_updated_after.return_value = False
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=SOURCE_OBJECT_NO_WILDCARD,
            last_modified_time=MOD_TIME_1,
        )

        operator.execute(None)
        mock_hook.return_value.rewrite.assert_not_called()

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_executes_with_is_older_than_with_true_cond(self, mock_hook):
        mock_hook.return_value.is_older_than.return_value = True
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=SOURCE_OBJECT_NO_WILDCARD,
            last_modified_time=MOD_TIME_1,
            maximum_modified_time=MOD_TIME_2,
            is_older_than=3600,
        )

        operator.execute(None)
        mock_hook.return_value.rewrite.assert_called_once_with(
            TEST_BUCKET, SOURCE_OBJECT_NO_WILDCARD, DESTINATION_BUCKET, SOURCE_OBJECT_NO_WILDCARD
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_more_than_1_wildcard(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_MULTIPLE_WILDCARDS,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=DESTINATION_OBJECT_PREFIX,
        )

        total_wildcards = operator.source_object.count(WILDCARD)

        error_msg = f"Only one wildcard '[*]' is allowed in source_object parameter. Found {total_wildcards}"

        with pytest.raises(AirflowException, match=error_msg):
            operator.execute(None)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_with_empty_destination_bucket(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_NO_WILDCARD,
            destination_bucket=None,
            destination_object=DESTINATION_OBJECT_PREFIX,
        )

        with mock.patch.object(operator.log, "warning") as mock_warn:
            operator.execute(None)
            mock_warn.assert_called_once_with(
                "destination_bucket is None. Defaulting it to source_bucket (%s)", TEST_BUCKET
            )
            assert operator.destination_bucket == operator.source_bucket

    # Tests the use of delimiter and source object as list
    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_executes_with_empty_source_objects(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID, source_bucket=TEST_BUCKET, source_objects=SOURCE_OBJECTS_NO_FILE
        )

        operator.execute(None)
        mock_hook.return_value.list.assert_called_once_with(
            TEST_BUCKET, prefix="", delimiter=None, match_glob=None
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_raises_exception_with_two_empty_list_inside_source_objects(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST
        operator = GCSToGCSOperator(
            task_id=TASK_ID, source_bucket=TEST_BUCKET, source_objects=SOURCE_OBJECTS_TWO_EMPTY_STRING
        )

        with pytest.raises(AirflowException, match="You can't have two empty strings inside source_object"):
            operator.execute(None)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_executes_with_single_item_in_source_objects(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID, source_bucket=TEST_BUCKET, source_objects=SOURCE_OBJECTS_SINGLE_FILE
        )
        operator.execute(None)
        mock_hook.return_value.list.assert_called_once_with(
            TEST_BUCKET, prefix=SOURCE_OBJECTS_SINGLE_FILE[0], delimiter=None, match_glob=None
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_executes_with_multiple_items_in_source_objects(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID, source_bucket=TEST_BUCKET, source_objects=SOURCE_OBJECTS_MULTIPLE_FILES
        )
        operator.execute(None)
        mock_hook.return_value.list.assert_has_calls(
            [
                mock.call(TEST_BUCKET, prefix="test_object/file1.txt", delimiter=None, match_glob=None),
                mock.call(TEST_BUCKET, prefix="test_object/file2.txt", delimiter=None, match_glob=None),
            ],
            any_order=True,
        )

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_executes_with_a_delimiter(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_objects=SOURCE_OBJECTS_NO_FILE,
            delimiter=DELIMITER,
        )
        operator.execute(None)
        mock_hook.return_value.list.assert_called_once_with(
            TEST_BUCKET, prefix="", delimiter=DELIMITER, match_glob=None
        )

    # COPY
    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_executes_with_delimiter_and_destination_object(self, mock_hook):
        mock_hook.return_value.list.side_effect = [[], [], [SOURCE_OBJECTS_LIST[2]]]
        mock_hook.return_value.exists.return_value = False

        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_objects=SOURCE_OBJECTS_LIST,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=DESTINATION_OBJECT,
            delimiter=DELIMITER,
        )

        operator.execute(None)
        mock_calls = [
            mock.call(
                TEST_BUCKET, SOURCE_OBJECTS_LIST[2], DESTINATION_BUCKET, DESTINATION_OBJECT + "file3.json"
            ),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_executes_with_different_delimiter_and_destination_object(self, mock_hook):
        mock_hook.return_value.list.side_effect = [[SOURCE_OBJECTS_LIST[0]], [SOURCE_OBJECTS_LIST[1]], []]
        mock_hook.return_value.exists.return_value = False

        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_objects=SOURCE_OBJECTS_LIST,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=DESTINATION_OBJECT,
            delimiter=".txt",
        )

        operator.execute(None)
        mock_calls = [
            mock.call(
                TEST_BUCKET, SOURCE_OBJECTS_LIST[0], DESTINATION_BUCKET, DESTINATION_OBJECT + "file1.txt"
            ),
            mock.call(
                TEST_BUCKET, SOURCE_OBJECTS_LIST[1], DESTINATION_BUCKET, DESTINATION_OBJECT + "file2.txt"
            ),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_executes_with_no_destination_bucket_and_no_destination_object(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST
        operator = GCSToGCSOperator(
            task_id=TASK_ID, source_bucket=TEST_BUCKET, source_objects=SOURCE_OBJECTS_LIST
        )
        operator.execute(None)
        mock_calls = [
            mock.call(TEST_BUCKET, "test_object/file1.txt", TEST_BUCKET, "test_object/file1.txt"),
            mock.call(TEST_BUCKET, "test_object/file2.txt", TEST_BUCKET, "test_object/file2.txt"),
            mock.call(TEST_BUCKET, "test_object/file3.json", TEST_BUCKET, "test_object/file3.json"),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_wc_with_last_modified_time_with_all_true_cond_no_file(self, mock_hook):
        mock_hook.return_value.list.return_value = SOURCE_OBJECTS_LIST
        mock_hook.return_value.is_updated_after.side_effect = [True, True, True]
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_objects=SOURCE_OBJECTS_NO_FILE,
            destination_bucket=DESTINATION_BUCKET,
            last_modified_time=MOD_TIME_1,
        )

        operator.execute(None)
        mock_calls_none = [
            mock.call(TEST_BUCKET, "test_object/file1.txt", DESTINATION_BUCKET, "test_object/file1.txt"),
            mock.call(TEST_BUCKET, "test_object/file2.txt", DESTINATION_BUCKET, "test_object/file2.txt"),
            mock.call(TEST_BUCKET, "test_object/file3.json", DESTINATION_BUCKET, "test_object/file3.json"),
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls_none)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_wildcard_with_replace_flag_false_with_destination_object(self, mock_hook):
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=SOURCE_OBJECT_WILDCARD_SUFFIX,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=DESTINATION_OBJECT_PREFIX,
            replace=False,
        )

        operator.execute(None)
        mock_calls = [
            mock.call(TEST_BUCKET, prefix="test_object", delimiter=""),
            mock.call(DESTINATION_BUCKET, prefix="foo/bar", delimiter="", match_glob=None),
        ]
        mock_hook.return_value.list.assert_has_calls(mock_calls)

    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_execute_source_object_required_flag_true(self, mock_hook):
        mock_hook.return_value.exists.return_value = False
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_objects=SOURCE_OBJECTS_SINGLE_FILE,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=DESTINATION_OBJECT_PREFIX,
            source_object_required=True,
        )

        with pytest.raises(
            AirflowException, match=f"{SOURCE_OBJECTS_SINGLE_FILE} does not exist in bucket {TEST_BUCKET}"
        ):
            operator.execute(None)

    @pytest.mark.parametrize(
        "existing_objects, source_object, match_glob, exact_match, expected_source_objects, "
        "expected_destination_objects",
        [
            (["source/foo.txt"], "source/foo.txt", None, True, ["source/foo.txt"], ["{prefix}/foo.txt"]),
            (["source/foo.txt"], "source/foo.txt", None, False, ["source/foo.txt"], ["{prefix}/foo.txt"]),
            (["source/foo.txt"], "source", None, False, ["source/foo.txt"], ["{prefix}/foo.txt"]),
            (["source/foo.txt"], "source/", None, False, ["source/foo.txt"], ["{prefix}/foo.txt"]),
            (["source/foo.txt"], "source/*", None, False, ["source/foo.txt"], ["{prefix}/foo.txt"]),
            (["source/foo.txt"], "source/foo.*", None, False, ["source/foo.txt"], ["{prefix}/txt"]),
            (["source/foo.txt"], "source/", "**/foo*", False, ["source/foo.txt"], ["{prefix}/foo.txt"]),
            (["source/foo.txt"], "source/", "**/foo.txt", False, ["source/foo.txt"], ["{prefix}/foo.txt"]),
            (
                ["source/foo.txt", "source/foo.txt.abc"],
                "source/foo.txt",
                None,
                True,
                ["source/foo.txt"],
                ["{prefix}/foo.txt"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc"],
                "source/foo.txt",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc"],
                "source",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc"],
                "source/",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc"],
                "source/*",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc"],
                "source/foo.*",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc"],
                ["{prefix}/txt", "{prefix}/txt.abc"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc"],
                "source/",
                "**/foo*",
                False,
                ["source/foo.txt", "source/foo.txt.abc"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc"],
                "source/",
                "**/foo.txt",
                False,
                ["source/foo.txt"],
                ["{prefix}/foo.txt"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                "source/foo.txt",
                None,
                True,
                ["source/foo.txt"],
                ["{prefix}/foo.txt"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                "source/foo.txt",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc", "{prefix}/foo.txt/subfolder/file.txt"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                "source",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc", "{prefix}/foo.txt/subfolder/file.txt"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                "source/",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc", "{prefix}/foo.txt/subfolder/file.txt"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                "source/*",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc", "{prefix}/foo.txt/subfolder/file.txt"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                "source/foo.*",
                None,
                False,
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                ["{prefix}/txt", "{prefix}/txt.abc", "{prefix}/txt/subfolder/file.txt"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                "source/",
                "**/foo*",
                False,
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                ["{prefix}/foo.txt", "{prefix}/foo.txt.abc", "{prefix}/foo.txt/subfolder/file.txt"],
            ),
            (
                ["source/foo.txt", "source/foo.txt.abc", "source/foo.txt/subfolder/file.txt"],
                "source/",
                "**/foo.txt",
                False,
                ["source/foo.txt"],
                ["{prefix}/foo.txt"],
            ),
        ],
    )
    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_copy_files_into_a_folder(
        self,
        mock_hook,
        existing_objects,
        source_object,
        match_glob,
        exact_match,
        expected_source_objects,
        expected_destination_objects,
    ):
        mock_hook.return_value.list.return_value = existing_objects
        operator = GCSToGCSOperator(
            task_id=TASK_ID,
            source_bucket=TEST_BUCKET,
            source_object=source_object,
            destination_bucket=DESTINATION_BUCKET,
            destination_object=DESTINATION_OBJECT_PREFIX + "/",
            exact_match=exact_match,
            match_glob=match_glob,
        )
        operator.execute(None)

        mock_calls = [
            mock.call(TEST_BUCKET, src, DESTINATION_BUCKET, dst.format(prefix=DESTINATION_OBJECT_PREFIX))
            for src, dst in zip(expected_source_objects, expected_destination_objects)
        ]
        mock_hook.return_value.rewrite.assert_has_calls(mock_calls)

    @pytest.mark.parametrize(
        ("source_objects", "destination_object", "inputs", "outputs"),
        (
            (
                SOURCE_OBJECTS_SINGLE_FILE,
                None,
                [Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_SINGLE_FILE[0])],
                [Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name=SOURCE_OBJECTS_SINGLE_FILE[0])],
            ),
            (
                SOURCE_OBJECTS_SINGLE_FILE,
                "target.txt",
                [Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_SINGLE_FILE[0])],
                [Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name="target.txt")],
            ),
            (
                SOURCE_OBJECTS_SINGLE_FILE,
                "target_pre",
                [Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_SINGLE_FILE[0])],
                [Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name="/")],
            ),
            (
                SOURCE_OBJECTS_SINGLE_FILE,
                "dir/",
                [Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_SINGLE_FILE[0])],
                [Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name="dir")],
            ),
            (
                SOURCE_OBJECTS_LIST,
                "",
                [
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_LIST[0]),
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_LIST[1]),
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_LIST[2]),
                ],
                [
                    Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name="/"),
                ],
            ),
            (
                [*SOURCE_OBJECTS_LIST, "dir/*"],
                "parent/pre_",
                [
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_LIST[0]),
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_LIST[1]),
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_LIST[2]),
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name="dir"),
                ],
                [Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name="parent")],
            ),
            (
                SOURCE_OBJECTS_NO_FILE,
                "no_ending_slash",
                [Dataset(namespace=f"gs://{TEST_BUCKET}", name="/")],
                [Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name="/")],
            ),
            (
                [
                    f"dir/{SOURCE_OBJECT_WILDCARD_PREFIX}",
                    f"dir/{SOURCE_OBJECT_WILDCARD_SUFFIX}",
                    f"dir/{SOURCE_OBJECT_WILDCARD_MIDDLE}",
                    f"dir/{SOURCE_OBJECT_WILDCARD_FILENAME}",
                    "dir/*",
                    "dir/",
                    "dir/pre_",
                ],
                "/",
                [
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name="dir"),
                ],
                [
                    Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name="/"),
                ],
            ),
            (
                ["", "dir/pre", SOURCE_OBJECTS_SINGLE_FILE[0]],
                DESTINATION_OBJECT,
                [
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name="/"),
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name="dir"),
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name=SOURCE_OBJECTS_SINGLE_FILE[0]),
                ],
                [
                    Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name=DESTINATION_OBJECT_PREFIX),
                ],
            ),
            (
                [
                    "",
                    "dir/",
                ],
                None,
                [
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name="/"),
                    Dataset(namespace=f"gs://{TEST_BUCKET}", name="dir"),
                ],
                [
                    Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name="/"),
                    Dataset(namespace=f"gs://{DESTINATION_BUCKET}", name="dir"),
                ],
            ),
        ),
        ids=(
            "single file without output",
            "single file with single file output",
            "single file with prefix output",
            "single file with dir output",
            "multiple file with empty output",
            "multiple file with prefix as output",
            "empty prefix with prefix as output",
            "directory + prefix or wildcard without output",
            "mixed prefixes and file paths with output dir",
            "empty prefix + directory without output",
        ),
    )
    @mock.patch("airflow.providers.google.cloud.transfers.gcs_to_gcs.GCSHook")
    def test_get_openlineage_facets_on_complete(
        self, mock_hook, source_objects, destination_object, inputs, outputs
    ):
        if source_objects and any(WILDCARD in obj for obj in source_objects):
            with pytest.warns(AirflowProviderDeprecationWarning, match="Usage of wildcard"):
                operator = GCSToGCSOperator(
                    task_id=TASK_ID,
                    source_bucket=TEST_BUCKET,
                    source_objects=source_objects,
                    destination_bucket=DESTINATION_BUCKET,
                    destination_object=destination_object,
                )
        else:
            operator = GCSToGCSOperator(
                task_id=TASK_ID,
                source_bucket=TEST_BUCKET,
                source_objects=source_objects,
                destination_bucket=DESTINATION_BUCKET,
                destination_object=destination_object,
            )

        operator.execute(None)

        lineage = operator.get_openlineage_facets_on_complete(None)
        assert len(lineage.inputs) == len(inputs)
        assert len(lineage.outputs) == len(outputs)
        assert all(element in lineage.inputs for element in inputs)
        assert all(element in inputs for element in lineage.inputs)
        assert all(element in lineage.outputs for element in outputs)
        assert all(element in outputs for element in lineage.outputs)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import os
from unittest import mock

import pytest
from google.cloud.bigquery import Row

from airflow.providers.google.cloud.transfers.bigquery_to_mssql import BigQueryToMsSqlOperator
from airflow.providers.microsoft.mssql.hooks.mssql import MsSqlHook

AIRFLOW_CONN_MSSQL_DEFAULT = "mssql://sa:airflow123@mssql:1433/"
TASK_ID = "test-bq-create-table-operator"
TEST_PROJECT_ID = "test-project"
TEST_DATASET = "test_dataset"
TEST_TABLE_ID = "Persons"
TEST_ROWS = [
    Row((0, "Airflow", "Apache", "1000 N West Street, Suite 1200", "Wilmington, NC, USA"), field_to_index={})
]


@pytest.mark.integration("mssql")
class TestBigQueryToMsSqlOperator:
    def setup_method(self):
        os.environ["AIRFLOW_CONN_MSSQL_DEFAULT"] = AIRFLOW_CONN_MSSQL_DEFAULT
        mssql_hook = MsSqlHook()
        mssql_conn = mssql_hook.get_conn()
        mssql_hook.set_autocommit(mssql_conn, True)
        self.mssql_cursor = mssql_conn.cursor()
        self.mssql_cursor.execute(f"""CREATE TABLE {TEST_TABLE_ID} (
                            PersonID int,
                            LastName varchar(255),
                            FirstName varchar(255),
                            Address varchar(255),
                            City varchar(255)
                            )""")

    def teardown_method(self):
        self.mssql_cursor.execute(f"DROP TABLE {TEST_TABLE_ID}")

    @mock.patch(
        "airflow.providers.google.cloud.transfers.bigquery_to_sql.BigQueryHook",
        return_value=mock.MagicMock(list_rows=mock.MagicMock(side_effect=[TEST_ROWS, []])),
    )
    def test_execute(self, mock_hook):
        operator = BigQueryToMsSqlOperator(
            task_id=TASK_ID,
            source_project_dataset_table=f"{TEST_PROJECT_ID}.{TEST_DATASET}.{TEST_TABLE_ID}",
            target_table_name=TEST_TABLE_ID,
            selected_fields=["PersonID", "LastName", "FirstName", "Address", "City"],
            replace=False,
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.return_value.list_rows.assert_has_calls(
            [
                mock.call(
                    dataset_id=TEST_DATASET,
                    table_id=TEST_TABLE_ID,
                    max_results=1000,
                    selected_fields=["PersonID", "LastName", "FirstName", "Address", "City"],
                    start_index=0,
                ),
                mock.call(
                    dataset_id=TEST_DATASET,
                    table_id=TEST_TABLE_ID,
                    max_results=1000,
                    selected_fields=["PersonID", "LastName", "FirstName", "Address", "City"],
                    start_index=1000,
                ),
            ],
            any_order=False,
        )
        self.mssql_cursor.execute(f"SELECT * FROM {TEST_TABLE_ID}")
        res = self.mssql_cursor.fetchone()
        assert res == TEST_ROWS[0].values()



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING

from attr import define, field

from airflow.providers.google import __version__ as provider_version

if TYPE_CHECKING:
    from openlineage.client.generated.base import RunFacet

try:
    try:
        from openlineage.client.generated.base import RunFacet
    except ImportError:  # Old OpenLineage client is used
        from openlineage.client.facet import BaseFacet as RunFacet  # type: ignore[assignment]

    @define
    class BigQueryJobRunFacet(RunFacet):
        """
        Facet that represents relevant statistics of bigquery run.

        :param cached: BigQuery caches query results. Rest of the statistics will not be provided for cached queries.
        :param billedBytes: How many bytes BigQuery bills for.
        :param properties: Full property tree of BigQUery run.
        """

        cached: bool
        billedBytes: int | None = field(default=None)
        properties: str | None = field(default=None)

        @staticmethod
        def _get_schema() -> str:
            return (
                "https://raw.githubusercontent.com/apache/airflow/"
                f"providers-google/{provider_version}/airflow/providers/google/"
                "openlineage/BigQueryJobRunFacet.json"
            )
except ImportError:  # OpenLineage is not available

    def create_no_op(*_, **__) -> None:
        """
        Create a no-op placeholder.

        This function creates and returns a None value, used as a placeholder when the OpenLineage client
        library is available. It represents an action that has no effect.
        """
        return None

    BigQueryJobRunFacet = create_no_op  # type: ignore[misc, assignment]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""This module contains a Google Cloud Storage to BigQuery operator."""

from __future__ import annotations

import json
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from google.api_core.exceptions import BadRequest, Conflict
from google.cloud.bigquery import (
    DEFAULT_RETRY,
    CopyJob,
    ExternalConfig,
    ExtractJob,
    LoadJob,
    QueryJob,
    SchemaField,
    UnknownJob,
)
from google.cloud.bigquery.table import EncryptionConfiguration, Table, TableReference

from airflow.configuration import conf
from airflow.exceptions import AirflowException
from airflow.providers.google.cloud.hooks.bigquery import BigQueryHook, BigQueryJob
from airflow.providers.google.cloud.hooks.gcs import GCSHook
from airflow.providers.google.cloud.links.bigquery import BigQueryTableLink
from airflow.providers.google.cloud.triggers.bigquery import BigQueryInsertJobTrigger
from airflow.providers.google.common.hooks.base_google import PROVIDE_PROJECT_ID
from airflow.providers.google.version_compat import BaseOperator
from airflow.utils.helpers import merge_dicts

if TYPE_CHECKING:
    from google.api_core.retry import Retry

    from airflow.utils.context import Context

ALLOWED_FORMATS = [
    "CSV",
    "NEWLINE_DELIMITED_JSON",
    "AVRO",
    "GOOGLE_SHEETS",
    "DATASTORE_BACKUP",
    "PARQUET",
    "ORC",
]


class GCSToBigQueryOperator(BaseOperator):
    """
    Loads files from Google Cloud Storage into BigQuery.

    The schema to be used for the BigQuery table may be specified in one of
    two ways. You may either directly pass the schema fields in, or you may
    point the operator to a Google Cloud Storage object name. The object in
    Google Cloud Storage must be a JSON file with the schema fields in it.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:GCSToBigQueryOperator`

    :param bucket: The bucket to load from. (templated)
    :param source_objects: String or List of Google Cloud Storage URIs to load from. (templated)
        If source_format is 'DATASTORE_BACKUP', the list must only contain a single URI.
    :param destination_project_dataset_table: The dotted
        ``(<project>.|<project>:)<dataset>.<table>`` BigQuery table to load data into.
        If ``<project>`` is not included, project will be the project defined in
        the connection json. (templated)
    :param schema_fields: If set, the schema field list as defined here:
        https://cloud.google.com/bigquery/docs/reference/v2/jobs#configuration.load
        Should not be set when source_format is 'DATASTORE_BACKUP'.
        Parameter must be defined if 'schema_object' is null and autodetect is False.
    :param schema_object: If set, a GCS object path pointing to a .json file that
        contains the schema for the table. (templated)
        Parameter must be defined if 'schema_fields' is null and autodetect is False.
    :param schema_object_bucket: [Optional] If set, the GCS bucket where the schema object
        template is stored. (templated) (Default: the value of ``bucket``)
    :param source_format: File format to export.
    :param compression: [Optional] The compression type of the data source.
        Possible values include GZIP and NONE.
        The default value is NONE.
        This setting is ignored for Google Cloud Bigtable,
        Google Cloud Datastore backups and Avro formats.
    :param create_disposition: The create disposition if the table doesn't exist.
    :param skip_leading_rows: The number of rows at the top of a CSV file that BigQuery
        will skip when loading the data.
        When autodetect is on, the behavior is the following:
        skip_leading_rows unspecified - Autodetect tries to detect headers in the first row.
        If they are not detected, the row is read as data. Otherwise, data is read starting
        from the second row.
        skip_leading_rows is 0 - Instructs autodetect that there are no headers and data
        should be read starting from the first row.
        skip_leading_rows = N > 0 - Autodetect skips N-1 rows and tries to detect headers
        in row N. If headers are not detected, row N is just skipped. Otherwise, row N is
        used to extract column names for the detected schema.
        Default value set to None so that autodetect option can detect schema fields.
    :param write_disposition: The write disposition if the table already exists.
    :param field_delimiter: The delimiter to use when loading from a CSV.
    :param max_bad_records: The maximum number of bad records that BigQuery can
        ignore when running the job.
    :param quote_character: The value that is used to quote data sections in a CSV file.
    :param ignore_unknown_values: [Optional] Indicates if BigQuery should allow
        extra values that are not represented in the table schema.
        If true, the extra values are ignored. If false, records with extra columns
        are treated as bad records, and if there are too many bad records, an
        invalid error is returned in the job result.
    :param allow_quoted_newlines: Whether to allow quoted newlines (true) or not (false).
    :param allow_jagged_rows: Accept rows that are missing trailing optional columns.
        The missing values are treated as nulls. If false, records with missing trailing
        columns are treated as bad records, and if there are too many bad records, an
        invalid error is returned in the job result. Only applicable to CSV, ignored
        for other formats.
    :param encoding: The character encoding of the data. See:
        https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs#configuration.query.tableDefinitions.(key).csvOptions.encoding
        https://cloud.google.com/bigquery/docs/reference/rest/v2/tables#externalDataConfiguration.csvOptions.encoding
    :param max_id_key: If set, the name of a column in the BigQuery table
        that's to be loaded. This will be used to select the MAX value from
        BigQuery after the load occurs. The results will be returned by the
        execute() command, which in turn gets stored in XCom for future
        operators to use. This can be helpful with incremental loads--during
        future executions, you can pick up from the max ID.
    :param schema_update_options: Allows the schema of the destination
        table to be updated as a side effect of the load job.
    :param src_fmt_configs: configure optional fields specific to the source format
    :param external_table: Flag to specify if the destination table should be
        a BigQuery external table. Default Value is False.
    :param time_partitioning: configure optional time partitioning fields i.e.
        partition by field, type and  expiration as per API specifications.
        Note that 'field' is not available in concurrency with
        dataset.table$partition.
    :param cluster_fields: Request that the result of this load be stored sorted
        by one or more columns. BigQuery supports clustering for both partitioned and
        non-partitioned tables. The order of columns given determines the sort order.
        Not applicable for external tables.
    :param autodetect: [Optional] Indicates if we should automatically infer the
        options and schema for CSV and JSON sources. (Default: ``True``).
        Parameter must be set to True if 'schema_fields' and 'schema_object' are undefined.
        It is suggested to set to True if table are create outside of Airflow.
        If autodetect is None and no schema is provided (neither via schema_fields
        nor a schema_object), assume the table already exists.
    :param encryption_configuration: [Optional] Custom encryption configuration (e.g., Cloud KMS keys).

        .. code-block:: python

            encryption_configuration = {
                "kmsKeyName": "projects/testp/locations/us/keyRings/test-kr/cryptoKeys/test-key",
            }
    :param location: [Optional] The geographic location of the job. Required except for US and EU.
        See details at https://cloud.google.com/bigquery/docs/locations#specifying_your_location
    :param impersonation_chain: Optional service account to impersonate using short-term
        credentials, or chained list of accounts required to get the access_token
        of the last account in the list, which will be impersonated in the request.
        If set as a string, the account must grant the originating account
        the Service Account Token Creator IAM role.
        If set as a sequence, the identities from the list must grant
        Service Account Token Creator IAM role to the directly preceding identity, with first
        account from the list granting this role to the originating account (templated).
    :param labels: [Optional] Labels for the BiqQuery table.
    :param description: [Optional] Description for the BigQuery table. This will only be used if the
        destination table is newly created. If the table already exists and a value different than the
        current description is provided, the job will fail.
    :param deferrable: Run operator in the deferrable mode
    :param force_delete: Force the destination table to be deleted if it already exists.
    """

    template_fields: Sequence[str] = (
        "bucket",
        "source_objects",
        "schema_object",
        "schema_object_bucket",
        "destination_project_dataset_table",
        "impersonation_chain",
        "src_fmt_configs",
    )
    template_ext: Sequence[str] = (".sql",)
    ui_color = "#f0eee4"
    operator_extra_links = (BigQueryTableLink(),)

    def __init__(
        self,
        *,
        bucket,
        source_objects,
        destination_project_dataset_table,
        schema_fields=None,
        schema_object=None,
        schema_object_bucket=None,
        source_format="CSV",
        compression="NONE",
        create_disposition="CREATE_IF_NEEDED",
        skip_leading_rows=None,
        write_disposition="WRITE_EMPTY",
        field_delimiter=",",
        max_bad_records=0,
        quote_character=None,
        ignore_unknown_values=False,
        allow_quoted_newlines=False,
        allow_jagged_rows=False,
        encoding="UTF-8",
        max_id_key=None,
        gcp_conn_id="google_cloud_default",
        schema_update_options=(),
        src_fmt_configs=None,
        external_table=False,
        time_partitioning=None,
        cluster_fields=None,
        autodetect=True,
        encryption_configuration=None,
        location=None,
        impersonation_chain: str | Sequence[str] | None = None,
        labels=None,
        description=None,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        result_retry: Retry = DEFAULT_RETRY,
        result_timeout: float | None = None,
        cancel_on_kill: bool = True,
        job_id: str | None = None,
        force_rerun: bool = True,
        reattach_states: set[str] | None = None,
        project_id: str = PROVIDE_PROJECT_ID,
        force_delete: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.hook: BigQueryHook | None = None
        self.configuration: dict[str, Any] = {}

        # GCS config
        if src_fmt_configs is None:
            src_fmt_configs = {}
        if time_partitioning is None:
            time_partitioning = {}
        self.bucket = bucket
        self.source_objects = source_objects
        self.schema_object = schema_object

        if schema_object_bucket is None:
            schema_object_bucket = bucket
        self.schema_object_bucket = schema_object_bucket

        # BQ config
        self.destination_project_dataset_table = destination_project_dataset_table
        self.project_id = project_id
        self.schema_fields = schema_fields
        if source_format.upper() not in ALLOWED_FORMATS:
            raise ValueError(
                f"{source_format} is not a valid source format. "
                f"Please use one of the following types: {ALLOWED_FORMATS}."
            )
        self.source_format = source_format.upper()
        self.compression = compression
        self.create_disposition = create_disposition
        self.skip_leading_rows = skip_leading_rows
        self.write_disposition = write_disposition
        self.field_delimiter = field_delimiter
        self.max_bad_records = max_bad_records
        self.quote_character = quote_character
        self.ignore_unknown_values = ignore_unknown_values
        self.allow_quoted_newlines = allow_quoted_newlines
        self.allow_jagged_rows = allow_jagged_rows
        self.external_table = external_table
        self.encoding = encoding

        self.max_id_key = max_id_key
        self.gcp_conn_id = gcp_conn_id

        self.schema_update_options = schema_update_options
        self.src_fmt_configs = src_fmt_configs
        self.time_partitioning = time_partitioning
        self.cluster_fields = cluster_fields
        self.autodetect = autodetect
        self.encryption_configuration = encryption_configuration
        self.location = location
        self.impersonation_chain = impersonation_chain

        self.labels = labels
        self.description = description

        self.job_id = job_id
        self.deferrable = deferrable
        self.result_retry = result_retry
        self.result_timeout = result_timeout
        self.force_rerun = force_rerun
        self.reattach_states: set[str] = reattach_states or set()
        self.cancel_on_kill = cancel_on_kill
        self.force_delete = force_delete

        self.source_uris: list[str] = []

    def _submit_job(
        self,
        hook: BigQueryHook,
        job_id: str,
    ) -> BigQueryJob:
        # Submit a new job without waiting for it to complete.
        return hook.insert_job(
            configuration=self.configuration,
            project_id=self.project_id or hook.project_id,
            location=self.location,
            job_id=job_id,
            timeout=self.result_timeout,
            retry=self.result_retry,
            nowait=True,
        )

    @staticmethod
    def _handle_job_error(job: BigQueryJob | UnknownJob) -> None:
        if job.error_result:
            raise AirflowException(f"BigQuery job {job.job_id} failed: {job.error_result}")

    def execute(self, context: Context):
        hook = BigQueryHook(
            gcp_conn_id=self.gcp_conn_id,
            location=self.location,
            impersonation_chain=self.impersonation_chain,
        )
        self.hook = hook
        self.source_format = self.source_format.upper()

        job_id = self.hook.generate_job_id(
            job_id=self.job_id,
            dag_id=self.dag_id,
            task_id=self.task_id,
            logical_date=context["logical_date"],
            configuration=self.configuration,
            force_rerun=self.force_rerun,
        )

        self.source_objects = (
            self.source_objects if isinstance(self.source_objects, list) else [self.source_objects]
        )
        self.source_uris = [f"gs://{self.bucket}/{source_object}" for source_object in self.source_objects]

        if not self.schema_fields:
            # Check for self.autodetect explicitly False. self.autodetect equal to None
            # entails we do not want to detect schema from files. Instead, it means we
            # rely on an already existing table's schema
            if not self.schema_object and self.autodetect is False:
                raise AirflowException(
                    "Table schema was not found. Neither schema object nor schema fields were specified"
                )
            if self.schema_object and self.source_format != "DATASTORE_BACKUP":
                gcs_hook = GCSHook(
                    gcp_conn_id=self.gcp_conn_id,
                    impersonation_chain=self.impersonation_chain,
                )
                self.schema_fields = json.loads(
                    gcs_hook.download(self.schema_object_bucket, self.schema_object).decode("utf-8")
                )
                self.log.info("Loaded fields from schema object: %s", self.schema_fields)
            else:
                self.schema_fields = None

        if self.external_table:
            self.log.info("Creating a new BigQuery table for storing data...")
            table_obj_api_repr = self._create_external_table()

            BigQueryTableLink.persist(
                context=context,
                dataset_id=table_obj_api_repr["tableReference"]["datasetId"],
                project_id=table_obj_api_repr["tableReference"]["projectId"],
                table_id=table_obj_api_repr["tableReference"]["tableId"],
            )
            if self.max_id_key:
                max_id = self._find_max_value_in_column()
                return max_id
        else:
            if self.force_delete:
                self.log.info("Deleting table %s", self.destination_project_dataset_table)
                hook.delete_table(table_id=self.destination_project_dataset_table)
            else:
                self.log.info("Using existing BigQuery table for storing data...")
            self.configuration = self._use_existing_table()

            try:
                self.log.info("Executing: %s", self.configuration)
                job: BigQueryJob | UnknownJob = self._submit_job(self.hook, job_id)
            except Conflict:
                # If the job already exists retrieve it
                job = self.hook.get_job(
                    project_id=self.project_id or self.hook.project_id,
                    location=self.location,
                    job_id=job_id,
                )
                if job.state not in self.reattach_states:
                    # Same job configuration, so we need force_rerun
                    raise AirflowException(
                        f"Job with id: {job_id} already exists and is in {job.state} state. If you "
                        f"want to force rerun it consider setting `force_rerun=True`."
                        f"Or, if you want to reattach in this scenario add {job.state} to `reattach_states`"
                    )
                # Job already reached state DONE
                if job.state == "DONE":
                    raise AirflowException("Job is already in state DONE. Can not reattach to this job.")

                # We are reattaching to a job
                self.log.info("Reattaching to existing Job in state %s", job.state)
                self._handle_job_error(job)

            job_types = {
                LoadJob._JOB_TYPE: ["sourceTable", "destinationTable"],
                CopyJob._JOB_TYPE: ["sourceTable", "destinationTable"],
                ExtractJob._JOB_TYPE: ["sourceTable"],
                QueryJob._JOB_TYPE: ["destinationTable"],
            }

            if self.hook.project_id:
                for job_type, tables_prop in job_types.items():
                    job_configuration = job.to_api_repr()["configuration"]
                    if job_type in job_configuration:
                        for table_prop in tables_prop:
                            if table_prop in job_configuration[job_type]:
                                table = job_configuration[job_type][table_prop]
                                persist_kwargs = {
                                    "context": context,
                                    "table_id": table,
                                }
                                if not isinstance(table, str):
                                    persist_kwargs["table_id"] = table["tableId"]
                                    persist_kwargs["dataset_id"] = table["datasetId"]
                                    persist_kwargs["project_id"] = table["projectId"]
                                BigQueryTableLink.persist(**persist_kwargs)

            self.job_id = job.job_id
            context["ti"].xcom_push(key="job_id", value=self.job_id)
            if self.deferrable:
                self.defer(
                    timeout=self.execution_timeout,
                    trigger=BigQueryInsertJobTrigger(
                        conn_id=self.gcp_conn_id,
                        job_id=self.job_id,
                        project_id=self.project_id or self.hook.project_id,
                        location=self.location or self.hook.location,
                        impersonation_chain=self.impersonation_chain,
                    ),
                    method_name="execute_complete",
                )
            else:
                job.result(timeout=self.result_timeout, retry=self.result_retry)
                self._handle_job_error(job)
                if self.max_id_key:
                    return self._find_max_value_in_column()

    def execute_complete(self, context: Context, event: dict[str, Any]):
        """
        Return immediately and relies on trigger to throw a success event. Callback for the trigger.

        Relies on trigger to throw an exception, otherwise it assumes execution was successful.
        """
        if event["status"] == "error":
            raise AirflowException(event["message"])
        self.log.info(
            "%s completed with response %s ",
            self.task_id,
            event["message"],
        )
        # Save job_id as an attribute to be later used by listeners
        self.job_id = event.get("job_id")
        return self._find_max_value_in_column()

    def _find_max_value_in_column(self):
        hook = BigQueryHook(
            gcp_conn_id=self.gcp_conn_id,
            location=self.location,
            impersonation_chain=self.impersonation_chain,
        )
        if self.max_id_key:
            self.log.info("Selecting the MAX value from BigQuery column %r...", self.max_id_key)
            select_command = (
                f"SELECT MAX({self.max_id_key}) AS max_value FROM {self.destination_project_dataset_table}"
            )
            self.configuration = {
                "query": {
                    "query": select_command,
                    "useLegacySql": False,
                    "schemaUpdateOptions": [],
                }
            }
            try:
                job_id = hook.insert_job(
                    configuration=self.configuration, project_id=self.project_id or hook.project_id
                )
                rows = list(hook.get_job(job_id=job_id, location=self.location).result())
            except BadRequest as e:
                if "Unrecognized name:" in e.message:
                    raise AirflowException(
                        f"Could not determine MAX value in column {self.max_id_key} "
                        f"since the default value of 'string_field_n' was set by BQ"
                    )
                raise AirflowException(e.message)
            if rows:
                for row in rows:
                    max_id = row[0] if row[0] else 0
                    self.log.info(
                        "Loaded BQ data with MAX value of column %s.%s: %s",
                        self.destination_project_dataset_table,
                        self.max_id_key,
                        max_id,
                    )
                    return str(max_id)
            else:
                raise RuntimeError(f"The {select_command} returned no rows!")

    def _create_external_table(self):
        external_config_api_repr = {
            "autodetect": self.autodetect,
            "sourceFormat": self.source_format,
            "sourceUris": self.source_uris,
            "compression": self.compression.upper(),
            "ignoreUnknownValues": self.ignore_unknown_values,
        }
        # if following fields are not specified in src_fmt_configs,
        # honor the top-level params for backward-compatibility
        backward_compatibility_configs = {
            "skipLeadingRows": self.skip_leading_rows,
            "fieldDelimiter": self.field_delimiter,
            "quote": self.quote_character,
            "allowQuotedNewlines": self.allow_quoted_newlines,
            "allowJaggedRows": self.allow_jagged_rows,
            "encoding": self.encoding,
        }
        src_fmt_to_param_mapping = {"CSV": "csvOptions", "GOOGLE_SHEETS": "googleSheetsOptions"}
        src_fmt_to_configs_mapping = {
            "csvOptions": [
                "allowJaggedRows",
                "allowQuotedNewlines",
                "fieldDelimiter",
                "skipLeadingRows",
                "quote",
                "encoding",
                "preserveAsciiControlCharacters",
                "columnNameCharacterMap",
            ],
            "googleSheetsOptions": ["skipLeadingRows"],
        }
        if self.source_format in src_fmt_to_param_mapping:
            valid_configs = src_fmt_to_configs_mapping[src_fmt_to_param_mapping[self.source_format]]
            self.src_fmt_configs = self._validate_src_fmt_configs(
                self.source_format, self.src_fmt_configs, valid_configs, backward_compatibility_configs
            )
            external_config_api_repr[src_fmt_to_param_mapping[self.source_format]] = self.src_fmt_configs

        external_config = ExternalConfig.from_api_repr(external_config_api_repr)
        if self.schema_fields:
            external_config.schema = [SchemaField.from_api_repr(f) for f in self.schema_fields]
        if self.max_bad_records:
            external_config.max_bad_records = self.max_bad_records

        # build table definition
        table = Table(
            table_ref=TableReference.from_string(self.destination_project_dataset_table, self.hook.project_id)
        )
        table.external_data_configuration = external_config
        if self.labels:
            table.labels = self.labels

        if self.description:
            table.description = self.description

        if self.encryption_configuration:
            table.encryption_configuration = EncryptionConfiguration.from_api_repr(
                self.encryption_configuration
            )
        table_obj_api_repr = table.to_api_repr()

        self.log.info("Creating external table: %s", self.destination_project_dataset_table)
        self.hook.create_table(
            table_resource=table_obj_api_repr,
            project_id=self.project_id or self.hook.project_id,
            location=self.location,
            exists_ok=True,
        )
        self.log.info("External table created successfully: %s", self.destination_project_dataset_table)
        return table_obj_api_repr

    def _use_existing_table(self):
        destination_project_id, destination_dataset, destination_table = self.hook.split_tablename(
            table_input=self.destination_project_dataset_table,
            default_project_id=self.hook.project_id,
            var_name="destination_project_dataset_table",
        )

        # bigquery also allows you to define how you want a table's schema to change
        # as a side effect of a load
        # for more details:
        # https://cloud.google.com/bigquery/docs/reference/rest/v2/jobs#configuration.load.schemaUpdateOptions
        allowed_schema_update_options = ["ALLOW_FIELD_ADDITION", "ALLOW_FIELD_RELAXATION"]
        if not set(allowed_schema_update_options).issuperset(set(self.schema_update_options)):
            raise ValueError(
                f"{self.schema_update_options} contains invalid schema update options. "
                f"Please only use one or more of the following options: {allowed_schema_update_options}"
            )

        self.configuration = {
            "load": {
                "autodetect": self.autodetect,
                "createDisposition": self.create_disposition,
                "destinationTable": {
                    "projectId": destination_project_id,
                    "datasetId": destination_dataset,
                    "tableId": destination_table,
                },
                "sourceFormat": self.source_format,
                "sourceUris": self.source_uris,
                "writeDisposition": self.write_disposition,
                "ignoreUnknownValues": self.ignore_unknown_values,
            },
        }
        self.time_partitioning = self._cleanse_time_partitioning(
            self.destination_project_dataset_table, self.time_partitioning
        )
        if self.time_partitioning:
            self.configuration["load"].update({"timePartitioning": self.time_partitioning})

        if self.cluster_fields:
            self.configuration["load"].update({"clustering": {"fields": self.cluster_fields}})

        if self.schema_fields:
            self.configuration["load"]["schema"] = {"fields": self.schema_fields}

        if self.schema_update_options:
            # To provide backward compatibility
            self.schema_update_options = list(self.schema_update_options or [])
            self.log.info("Adding experimental 'schemaUpdateOptions': %s", self.schema_update_options)
            self.configuration["load"]["schemaUpdateOptions"] = self.schema_update_options

        if self.max_bad_records:
            self.configuration["load"]["maxBadRecords"] = self.max_bad_records

        if self.encryption_configuration:
            self.configuration["load"]["destinationEncryptionConfiguration"] = self.encryption_configuration

        if self.labels or self.description:
            self.configuration["load"].update({"destinationTableProperties": {}})
            if self.labels:
                self.configuration["load"]["destinationTableProperties"]["labels"] = self.labels
            if self.description:
                self.configuration["load"]["destinationTableProperties"]["description"] = self.description

        src_fmt_to_configs_mapping = {
            "CSV": [
                "allowJaggedRows",
                "allowQuotedNewlines",
                "autodetect",
                "fieldDelimiter",
                "skipLeadingRows",
                "ignoreUnknownValues",
                "nullMarker",
                "quote",
                "encoding",
                "preserveAsciiControlCharacters",
                "columnNameCharacterMap",
            ],
            "DATASTORE_BACKUP": ["projectionFields"],
            "NEWLINE_DELIMITED_JSON": ["autodetect", "ignoreUnknownValues"],
            "PARQUET": ["autodetect", "ignoreUnknownValues"],
            "AVRO": ["useAvroLogicalTypes"],
            "ORC": ["autodetect"],
        }

        valid_configs = src_fmt_to_configs_mapping[self.source_format]

        # if following fields are not specified in src_fmt_configs,
        # honor the top-level params for backward-compatibility
        backward_compatibility_configs = {
            "skipLeadingRows": self.skip_leading_rows,
            "fieldDelimiter": self.field_delimiter,
            "ignoreUnknownValues": self.ignore_unknown_values,
            "quote": self.quote_character,
            "allowQuotedNewlines": self.allow_quoted_newlines,
            "encoding": self.encoding,
        }

        self.src_fmt_configs = self._validate_src_fmt_configs(
            self.source_format, self.src_fmt_configs, valid_configs, backward_compatibility_configs
        )

        self.configuration["load"].update(self.src_fmt_configs)

        if self.allow_jagged_rows:
            self.configuration["load"]["allowJaggedRows"] = self.allow_jagged_rows
        return self.configuration

    def _validate_src_fmt_configs(
        self,
        source_format: str,
        src_fmt_configs: dict,
        valid_configs: list[str],
        backward_compatibility_configs: dict | None = None,
    ) -> dict:
        """
        Validate the given src_fmt_configs against a valid configuration for the source format.

        Adds the backward compatibility config to the src_fmt_configs.

        :param source_format: File format to export.
        :param src_fmt_configs: Configure optional fields specific to the source format.
        :param valid_configs: Valid configuration specific to the source format
        :param backward_compatibility_configs: The top-level params for backward-compatibility
        """
        if backward_compatibility_configs is None:
            backward_compatibility_configs = {}

        for k, v in backward_compatibility_configs.items():
            if k not in src_fmt_configs and k in valid_configs:
                src_fmt_configs[k] = v

        for k in src_fmt_configs:
            if k not in valid_configs:
                raise ValueError(f"{k} is not a valid src_fmt_configs for type {source_format}.")

        return src_fmt_configs

    def _cleanse_time_partitioning(
        self, destination_dataset_table: str | None, time_partitioning_in: dict | None
    ) -> dict:  # if it is a partitioned table ($ is in the table name) add partition load option
        if time_partitioning_in is None:
            time_partitioning_in = {}

        time_partitioning_out = {}
        if destination_dataset_table and "$" in destination_dataset_table:
            time_partitioning_out["type"] = "DAY"
        time_partitioning_out.update(time_partitioning_in)
        return time_partitioning_out

    def on_kill(self) -> None:
        if self.job_id and self.cancel_on_kill:
            self.hook.cancel_job(job_id=self.job_id, location=self.location)  # type: ignore[union-attr]
        else:
            self.log.info("Skipping to cancel job: %s.%s", self.location, self.job_id)

    def get_openlineage_facets_on_complete(self, task_instance):
        """Implement on_complete as we will include final BQ job id."""
        from airflow.providers.common.compat.openlineage.facet import (
            Dataset,
            ExternalQueryRunFacet,
            Identifier,
            SymlinksDatasetFacet,
        )
        from airflow.providers.google.cloud.openlineage.utils import (
            WILDCARD,
            extract_ds_name_from_gcs_path,
            get_facets_from_bq_table,
            get_identity_column_lineage_facet,
        )
        from airflow.providers.openlineage.extractors import OperatorLineage

        if not self.hook:
            self.hook = BigQueryHook(
                gcp_conn_id=self.gcp_conn_id,
                location=self.location,
                impersonation_chain=self.impersonation_chain,
            )

        project_id = self.project_id or self.hook.project_id
        table_object = self.hook.get_client(project_id).get_table(self.destination_project_dataset_table)

        output_dataset_facets = get_facets_from_bq_table(table_object)

        source_objects = (
            self.source_objects if isinstance(self.source_objects, list) else [self.source_objects]
        )
        input_dataset_facets = {}
        if "schema" in output_dataset_facets:
            input_dataset_facets["schema"] = output_dataset_facets["schema"]

        input_datasets = []
        for blob in sorted(source_objects):
            additional_facets = {}

            if WILDCARD in blob:
                # For path with wildcard we attach a symlink with unmodified path.
                additional_facets = {
                    "symlink": SymlinksDatasetFacet(
                        identifiers=[Identifier(namespace=f"gs://{self.bucket}", name=blob, type="file")]
                    ),
                }

            dataset = Dataset(
                namespace=f"gs://{self.bucket}",
                name=extract_ds_name_from_gcs_path(blob),
                facets=merge_dicts(input_dataset_facets, additional_facets),
            )
            input_datasets.append(dataset)

        output_dataset = Dataset(
            namespace="bigquery",
            name=str(table_object.reference),
            facets={
                **output_dataset_facets,
                **get_identity_column_lineage_facet(
                    dest_field_names=[field.name for field in table_object.schema],
                    input_datasets=input_datasets,
                ),
            },
        )

        run_facets = {}
        if self.job_id:
            run_facets = {
                "externalQuery": ExternalQueryRunFacet(externalQueryId=self.job_id, source="bigquery"),
            }

        return OperatorLineage(inputs=input_datasets, outputs=[output_dataset], run_facets=run_facets)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import os
from datetime import datetime

# [START howto_operator_import_protobuf_obj]
from google.cloud.logging_v2.types import LogSink
from google.protobuf.field_mask_pb2 import FieldMask

# [END howto_operator_import_protobuf_obj]
from airflow.models.dag import DAG
from airflow.providers.google.cloud.operators.cloud_logging_sink import (
    CloudLoggingCreateSinkOperator,
    CloudLoggingDeleteSinkOperator,
    CloudLoggingListSinksOperator,
    CloudLoggingUpdateSinkOperator,
)

PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT", "default")
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")

SINK_NAME = "example-airflow-test-sink"
CONN_ID = "google_cloud_default"

with DAG(
    dag_id="google_cloud_logging_sink",
    schedule="@once",
    start_date=datetime(2024, 1, 1),
    catchup=False,
    tags=["example", "gcp", "cloud-logging"],
) as dag:
    # [START howto_operator_cloud_logging_create_sink_native_obj]
    create_sink = CloudLoggingCreateSinkOperator(
        task_id="create_sink",
        project_id=PROJECT_ID,
        sink_config={
            "name": SINK_NAME,
            "destination": "storage.googleapis.com/test-log-sink-af",
            "description": "Create with full sink_config",
            "filter": "severity>=INFO",
            "disabled": False,
            "exclusions": [
                {
                    "name": "exclude-debug",
                    "description": "Skip debug logs",
                    "filter": "severity=DEBUG",
                    "disabled": True,
                },
                {
                    "name": "exclude-cloudsql",
                    "description": "Skip CloudSQL logs",
                    "filter": 'resource.type="cloudsql_database"',
                    "disabled": False,
                },
            ],
        },
        gcp_conn_id=CONN_ID,
    )
    # [END howto_operator_cloud_logging_create_sink_native_obj]

    # [START howto_operator_cloud_logging_update_sink_protobuf_obj]
    update_sink_config = CloudLoggingUpdateSinkOperator(
        task_id="update_sink_config",
        sink_name=SINK_NAME,
        project_id=PROJECT_ID,
        sink_config=LogSink(
            {
                "description": "Update #1: GCE logs only",
                "filter": 'resource.type="gce_instance"',
                "disabled": False,
            }
        ),
        update_mask=FieldMask(paths=["description", "filter", "disabled"]),
        unique_writer_identity=True,
        gcp_conn_id=CONN_ID,
    )
    # [END howto_operator_cloud_logging_update_sink_protobuf_obj]

    # [START howto_operator_cloud_logging_list_sinks]
    list_sinks_after = CloudLoggingListSinksOperator(
        task_id="list_sinks_after_update",
        project_id=PROJECT_ID,
        gcp_conn_id=CONN_ID,
    )
    # [END howto_operator_cloud_logging_list_sinks]

    # [START howto_operator_cloud_logging_delete_sink]
    delete_sink = CloudLoggingDeleteSinkOperator(
        task_id="delete_sink",
        sink_name=SINK_NAME,
        project_id=PROJECT_ID,
        gcp_conn_id=CONN_ID,
    )
    # [END howto_operator_cloud_logging_delete_sink]

    (create_sink >> update_sink_config >> list_sinks_after >> delete_sink)

    from tests_common.test_utils.watcher import watcher

    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that shows how to use GoogleAdsToGcsOperator.

In order to run this test, make sure you followed steps:
1. In your GCP project create a service account that will be used to operate on Google Ads.
The name should be in format `google-ads-service-account@{PROJECT_ID}.iam.gserviceaccount.com`
2. Generate a key for this service account and store it in the Secret Manager
under the name `google_ads_service_account_key`.
3. Give this service account Editor permissions.
4. Make sure Google Ads API is enabled in your GCP project.
5. Login to https://ads.google.com
6. In the Admin section go to Access and Security and give your GCP service account Admin permissions.
7. Store values of your developer token and client ID to Secret Manager under names `google_ads_client_id`
and `google_ads_developer_token`.
"""

from __future__ import annotations

import json
import logging
import os
from datetime import datetime

from google.cloud.exceptions import NotFound

try:
    from airflow.sdk import task
except ImportError:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.dag import DAG
from airflow.providers.google.ads.operators.ads import GoogleAdsListAccountsOperator
from airflow.providers.google.ads.transfers.ads_to_gcs import GoogleAdsToGcsOperator
from airflow.providers.google.cloud.hooks.secret_manager import GoogleCloudSecretManagerHook
from airflow.providers.google.cloud.operators.gcs import GCSCreateBucketOperator, GCSDeleteBucketOperator

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from tests_common.test_utils.api_client_helpers import create_airflow_connection, delete_airflow_connection

# [START howto_google_ads_env_variables]
ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_PROJECT_ID", "default")
API_VERSION = "v20"

DAG_ID = "google_ads"

GOOGLE_ADS_CLIENT_ID = "google_ads_client_id"
GOOGLE_ADS_SERVICE_ACCOUNT_KEY = "google_ads_service_account_key"
GOOGLE_ADS_DEVELOPER_TOKEN = "google_ads_developer_token"

BUCKET_NAME = f"bucket_ads_{ENV_ID}"
GCS_OBJ_PATH = f"gs://{BUCKET_NAME}/google-ads-api-results.csv"
GCS_ACCOUNTS_CSV = "accounts.csv"
QUERY = """
    SELECT
        segments.date,
        customer.id,
        campaign.id,
        ad_group.id,
        ad_group_ad.ad.id,
        metrics.impressions,
        metrics.clicks,
        metrics.conversions,
        metrics.all_conversions,
        metrics.cost_micros
    FROM
        ad_group_ad
    WHERE
        segments.date >= '2020-02-01'
        AND segments.date <= '2020-02-29'
    """
CONNECTION_GLOUD_ID = f"connection_cloud_{DAG_ID}_{ENV_ID}"
CONNECTION_ADS_ID = "google_ads_default"
CONNECTION_TYPE = "google_cloud_platform"

FIELDS_TO_EXTRACT = [
    "segments.date.value",
    "customer.id.value",
    "campaign.id.value",
    "ad_group.id.value",
    "ad_group_ad.ad.id.value",
    "metrics.impressions.value",
    "metrics.clicks.value",
    "metrics.conversions.value",
    "metrics.all_conversions.value",
    "metrics.cost_micros.value",
]
# [END howto_google_ads_env_variables]

log = logging.getLogger(__name__)


def get_secret(secret_id: str) -> str:
    hook = GoogleCloudSecretManagerHook()
    if hook.secret_exists(secret_id=secret_id):
        return hook.access_secret(secret_id=secret_id).payload.data.decode()
    raise NotFound("The secret '%s' not found", secret_id)


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
    tags=["example", "ads"],
    render_template_as_native_obj=True,
) as dag:

    @task
    def get_google_ads_client_id():
        return get_secret(secret_id=GOOGLE_ADS_CLIENT_ID).strip()

    get_google_ads_client_id_task = get_google_ads_client_id()

    @task
    def get_google_ads_service_account_key():
        return get_secret(secret_id=GOOGLE_ADS_SERVICE_ACCOUNT_KEY)

    get_google_ads_service_account_key_task = get_google_ads_service_account_key()

    @task
    def get_google_ads_developer_token():
        return get_secret(secret_id=GOOGLE_ADS_DEVELOPER_TOKEN).strip()

    get_google_ads_developer_token_task = get_google_ads_developer_token()

    @task
    def create_connection_gcloud_for_ads(connection_id: str, key) -> None:
        conn_extra_json = json.dumps(
            {
                "keyfile_dict": key,
                "project": PROJECT_ID,
                "scope": "https://www.googleapis.com/auth/adwords, https://www.googleapis.com/auth/cloud-platform",
            }
        )
        create_airflow_connection(
            connection_id=connection_id,
            connection_conf={"conn_type": CONNECTION_TYPE, "extra": conn_extra_json},
        )

    create_connection_gcloud_for_ads = create_connection_gcloud_for_ads(
        connection_id=CONNECTION_GLOUD_ID, key=get_google_ads_service_account_key_task
    )

    @task
    def create_connection_ads(connection_id: str, token) -> None:
        conn_extra_json = json.dumps(
            {
                "google_ads_client": {
                    "developer_token": token,
                    # this parameter is required to be not None, but the actual content will be overwritten, so can be some dummy string
                    "json_key_file_path": "some_string",
                    "impersonated_email": f"google-ads-service-account@{PROJECT_ID}.iam.gserviceaccount.com",
                    "use_proto_plus": False,
                },
                "project": PROJECT_ID,
                "scope": "https://www.googleapis.com/auth/adwords, https://www.googleapis.com/auth/cloud-platform",
            }
        )
        create_airflow_connection(
            connection_id=connection_id,
            connection_conf={"conn_type": CONNECTION_TYPE, "extra": conn_extra_json},
        )

    create_connection_ads_task = create_connection_ads(
        connection_id=CONNECTION_ADS_ID, token=get_google_ads_developer_token_task
    )

    create_bucket = GCSCreateBucketOperator(
        task_id="create_bucket",
        bucket_name=BUCKET_NAME,
        project_id=PROJECT_ID,
        gcp_conn_id=CONNECTION_GLOUD_ID,
    )

    # [START howto_google_ads_to_gcs_operator]
    run_operator = GoogleAdsToGcsOperator(
        client_ids=[get_google_ads_client_id_task],
        query=QUERY,
        attributes=FIELDS_TO_EXTRACT,
        obj=GCS_OBJ_PATH,
        bucket=BUCKET_NAME,
        task_id="run_operator",
        api_version=API_VERSION,
        gcp_conn_id=CONNECTION_GLOUD_ID,
    )
    # [END howto_google_ads_to_gcs_operator]

    # [START howto_ads_list_accounts_operator]
    list_accounts = GoogleAdsListAccountsOperator(
        task_id="list_accounts",
        bucket=BUCKET_NAME,
        object_name=GCS_ACCOUNTS_CSV,
        api_version=API_VERSION,
        gcp_conn_id=CONNECTION_GLOUD_ID,
    )
    # [END howto_ads_list_accounts_operator]

    delete_bucket = GCSDeleteBucketOperator(
        task_id="delete_bucket",
        bucket_name=BUCKET_NAME,
        gcp_conn_id=CONNECTION_GLOUD_ID,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    @task(task_id="delete_connection_gloud")
    def delete_connection_gloud(connection_id: str) -> None:
        delete_airflow_connection(connection_id=connection_id)

    delete_connection_gloud_task = delete_connection_gloud(connection_id=CONNECTION_GLOUD_ID)

    @task(task_id="delete_connection_ads")
    def delete_connection_ads(connection_id: str) -> None:
        delete_airflow_connection(connection_id=connection_id)

    delete_connection_ads_task = delete_connection_ads(connection_id=CONNECTION_ADS_ID)

    (
        # TEST SETUP
        [
            get_google_ads_client_id_task,
            get_google_ads_service_account_key_task,
            get_google_ads_developer_token_task,
        ]
        >> create_connection_gcloud_for_ads  # type: ignore
        >> create_connection_ads_task
        >> create_bucket
        # TEST BODY
        >> run_operator
        >> list_accounts
        # TEST TEARDOWN
        >> delete_bucket
        >> [delete_connection_gloud_task, delete_connection_ads_task]
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()


from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
Example Airflow DAG that performs query in a Cloud SQL instance with SSL support.
"""

from __future__ import annotations

import base64
import json
import logging
import os
import random
import string
from collections.abc import Iterable
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import Any

from googleapiclient import discovery

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
from airflow.models.dag import DAG
from airflow.providers.google.cloud.hooks.cloud_sql import CloudSQLHook
from airflow.providers.google.cloud.hooks.secret_manager import GoogleCloudSecretManagerHook
from airflow.providers.google.cloud.operators.cloud_sql import (
    CloudSQLCreateInstanceDatabaseOperator,
    CloudSQLCreateInstanceOperator,
    CloudSQLDeleteInstanceOperator,
    CloudSQLExecuteQueryOperator,
)

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.google import DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
from tests_common.test_utils.api_client_helpers import create_airflow_connection, delete_airflow_connection

ENV_ID = os.environ.get("SYSTEM_TESTS_ENV_ID", "default")
PROJECT_ID = os.environ.get("SYSTEM_TESTS_GCP_PROJECT") or DEFAULT_GCP_SYSTEM_TEST_PROJECT_ID
DAG_ID = "cloudsql_query_ssl"
REGION = "us-central1"
HOME_DIR = Path.home()

COMPOSER_ENVIRONMENT = os.environ.get("COMPOSER_ENVIRONMENT", "")


def run_in_composer():
    return bool(COMPOSER_ENVIRONMENT)


CLOUD_SQL_INSTANCE_NAME_TEMPLATE = f"{ENV_ID}-{DAG_ID}".replace("_", "-")
CLOUD_SQL_INSTANCE_CREATE_BODY_TEMPLATE: dict[str, Any] = {
    "name": CLOUD_SQL_INSTANCE_NAME_TEMPLATE,
    "settings": {
        "tier": "db-custom-1-3840",
        "dataDiskSizeGb": 30,
        "pricingPlan": "PER_USE",
        "ipConfiguration": {},
    },
    # For using a different database version please check the link below.
    # https://cloud.google.com/sql/docs/mysql/admin-api/rest/v1/SqlDatabaseVersion
    "databaseVersion": "1.2.3",
    "region": REGION,
    "ipConfiguration": {
        "ipv4Enabled": True,
        "requireSsl": True,
        "authorizedNetworks": [
            {"value": "0.0.0.0/0"},
        ],
    },
}

DB_PROVIDERS: Iterable[dict[str, str]] = (
    {
        "database_type": "postgres",
        "port": "5432",
        "database_version": "POSTGRES_15",
        "cloud_sql_instance_name": f"{CLOUD_SQL_INSTANCE_NAME_TEMPLATE}-postgres",
    },
    {
        "database_type": "mysql",
        "port": "3306",
        "database_version": "MYSQL_8_0",
        "cloud_sql_instance_name": f"{CLOUD_SQL_INSTANCE_NAME_TEMPLATE}-mysql",
    },
)


def ip_configuration() -> dict[str, Any]:
    """Generates an ip configuration for a CloudSQL instance creation body"""
    if run_in_composer():
        # Use connection to Cloud SQL instance via Private IP within the Cloud Composer's network.
        return {
            "ipv4Enabled": True,
            "requireSsl": False,
            "sslMode": "ENCRYPTED_ONLY",
            "enablePrivatePathForGoogleCloudServices": True,
            "privateNetwork": f"projects/{PROJECT_ID}/global/networks/default",
        }
    # Use connection to Cloud SQL instance via Public IP from anywhere (mask 0.0.0.0/0).
    # Consider specifying your network mask
    # for allowing requests only from the trusted sources, not from anywhere.
    return {
        "ipv4Enabled": True,
        "requireSsl": False,
        "sslMode": "ENCRYPTED_ONLY",
        "authorizedNetworks": [
            {"value": "0.0.0.0/0"},
        ],
    }


def cloud_sql_instance_create_body(database_provider: dict[str, Any]) -> dict[str, Any]:
    """Generates a CloudSQL instance creation body"""
    create_body: dict[str, Any] = deepcopy(CLOUD_SQL_INSTANCE_CREATE_BODY_TEMPLATE)
    create_body["name"] = database_provider["cloud_sql_instance_name"]
    create_body["databaseVersion"] = database_provider["database_version"]
    create_body["settings"]["ipConfiguration"] = ip_configuration()
    return create_body


CLOUD_SQL_DATABASE_NAME = "test_db"
CLOUD_SQL_USER = "test_user"
CLOUD_SQL_PASSWORD = "JoxHlwrPzwch0gz9"
CLOUD_SQL_IP_ADDRESS = "127.0.0.1"
CLOUD_SQL_PUBLIC_PORT = 5432


def cloud_sql_database_create_body(instance: str) -> dict[str, Any]:
    """Generates a CloudSQL database creation body"""
    return {
        "instance": instance,
        "name": CLOUD_SQL_DATABASE_NAME,
        "project": PROJECT_ID,
    }


CLOUD_SQL_INSTANCE_NAME = ""
DATABASE_TYPE = ""  # "postgres|mysql|mssql"

# [START howto_operator_cloudsql_query_connections]
# Connect directly via TCP (SSL)
CONNECTION_PUBLIC_TCP_SSL_KWARGS = {
    "conn_type": "gcpcloudsql",
    "login": CLOUD_SQL_USER,
    "password": CLOUD_SQL_PASSWORD,
    "host": CLOUD_SQL_IP_ADDRESS,
    "port": CLOUD_SQL_PUBLIC_PORT,
    "schema": CLOUD_SQL_DATABASE_NAME,
    "extra": {
        "database_type": DATABASE_TYPE,
        "project_id": PROJECT_ID,
        "location": REGION,
        "instance": CLOUD_SQL_INSTANCE_NAME,
        "use_proxy": "False",
        "use_ssl": "True",
    },
}
# [END howto_operator_cloudsql_query_connections]

CONNECTION_PUBLIC_TCP_SSL_ID = f"{DAG_ID}_{ENV_ID}_tcp_ssl"

SQL = [
    "CREATE TABLE IF NOT EXISTS TABLE_TEST (I INTEGER)",
    "CREATE TABLE IF NOT EXISTS TABLE_TEST (I INTEGER)",
    "INSERT INTO TABLE_TEST VALUES (0)",
    "CREATE TABLE IF NOT EXISTS TABLE_TEST2 (I INTEGER)",
    "DROP TABLE TABLE_TEST",
    "DROP TABLE TABLE_TEST2",
]

SSL_PATH = f"/{DAG_ID}/{ENV_ID}"
SSL_LOCAL_PATH_PREFIX = "/tmp"
SSL_COMPOSER_PATH_PREFIX = "/home/airflow/gcs/data"
# [START howto_operator_cloudsql_query_connections_env]

# The connections below are created using one of the standard approaches - via environment
# variables named AIRFLOW_CONN_* . The connections can also be created in the database
# of AIRFLOW (using command line or UI).

postgres_kwargs = {
    "user": "user",
    "password": "password",
    "public_ip": "public_ip",
    "public_port": "public_port",
    "database": "database",
    "project_id": "project_id",
    "location": "location",
    "instance": "instance",
    "client_cert_file": "client_cert_file",
    "client_key_file": "client_key_file",
    "server_ca_file": "server_ca_file",
}

# Postgres: connect directly via TCP (SSL)
os.environ["AIRFLOW_CONN_PUBLIC_POSTGRES_TCP_SSL"] = (
    "gcpcloudsql://{user}:{password}@{public_ip}:{public_port}/{database}?"
    "database_type=postgres&"
    "project_id={project_id}&"
    "location={location}&"
    "instance={instance}&"
    "use_proxy=False&"
    "use_ssl=True&"
    "sslcert={client_cert_file}&"
    "sslkey={client_key_file}&"
    "sslrootcert={server_ca_file}".format(**postgres_kwargs)
)

mysql_kwargs = {
    "user": "user",
    "password": "password",
    "public_ip": "public_ip",
    "public_port": "public_port",
    "database": "database",
    "project_id": "project_id",
    "location": "location",
    "instance": "instance",
    "client_cert_file": "client_cert_file",
    "client_key_file": "client_key_file",
    "server_ca_file": "server_ca_file",
}

# MySQL: connect directly via TCP (SSL)
os.environ["AIRFLOW_CONN_PUBLIC_MYSQL_TCP_SSL"] = (
    "gcpcloudsql://{user}:{password}@{public_ip}:{public_port}/{database}?"
    "database_type=mysql&"
    "project_id={project_id}&"
    "location={location}&"
    "instance={instance}&"
    "use_proxy=False&"
    "use_ssl=True&"
    "sslcert={client_cert_file}&"
    "sslkey={client_key_file}&"
    "sslrootcert={server_ca_file}".format(**mysql_kwargs)
)
# [END howto_operator_cloudsql_query_connections_env]

log = logging.getLogger(__name__)

with DAG(
    dag_id=DAG_ID,
    start_date=datetime(2021, 1, 1),
    schedule="@once",
    catchup=False,
    tags=["example", "cloudsql", "postgres"],
) as dag:
    for db_provider in DB_PROVIDERS:
        database_type: str = db_provider["database_type"]
        cloud_sql_instance_name: str = db_provider["cloud_sql_instance_name"]

        create_cloud_sql_instance = CloudSQLCreateInstanceOperator(
            task_id=f"create_cloud_sql_instance_{database_type}",
            project_id=PROJECT_ID,
            instance=cloud_sql_instance_name,
            body=cloud_sql_instance_create_body(database_provider=db_provider),
        )

        create_database = CloudSQLCreateInstanceDatabaseOperator(
            task_id=f"create_database_{database_type}",
            body=cloud_sql_database_create_body(instance=cloud_sql_instance_name),
            instance=cloud_sql_instance_name,
        )

        @task(task_id=f"create_user_{database_type}")
        def create_user(instance: str) -> None:
            with discovery.build("sqladmin", "v1beta4") as service:
                request = service.users().insert(
                    project=PROJECT_ID,
                    instance=instance,
                    body={
                        "name": CLOUD_SQL_USER,
                        "password": CLOUD_SQL_PASSWORD,
                    },
                )
                request.execute()
            return None

        create_user_task = create_user(instance=cloud_sql_instance_name)

        @task(task_id=f"get_ip_address_{database_type}")
        def get_ip_address(instance: str) -> str | None:
            """Returns a Cloud SQL instance IP address.

            If the test is running in Cloud Composer, the Private IP address is used, otherwise Public IP."""
            with discovery.build("sqladmin", "v1beta4") as service:
                request = service.connect().get(
                    project=PROJECT_ID,
                    instance=instance,
                    fields="ipAddresses",
                )
                response = request.execute()
                for ip_item in response.get("ipAddresses", []):
                    if run_in_composer():
                        if ip_item["type"] == "PRIVATE":
                            return ip_item["ipAddress"]
                    else:
                        if ip_item["type"] == "PRIMARY":
                            return ip_item["ipAddress"]
                return None

        get_ip_address_task = get_ip_address(instance=cloud_sql_instance_name)

        conn_id = f"{CONNECTION_PUBLIC_TCP_SSL_ID}_{database_type}"

        @task(task_id=f"create_connection_{database_type}")
        def create_connection(
            connection_id: str, instance: str, db_type: str, ip_address: str, port: str
        ) -> str | None:
            connection: dict[str, Any] = deepcopy(CONNECTION_PUBLIC_TCP_SSL_KWARGS)
            connection["extra"]["instance"] = instance
            connection["host"] = ip_address
            connection["extra"]["database_type"] = db_type
            connection["port"] = port
            connection["extra"] = json.dumps(connection["extra"])
            create_airflow_connection(
                connection_id=connection_id,
                connection_conf=connection,
            )
            return connection_id

        create_connection_task = create_connection(
            connection_id=conn_id,
            instance=cloud_sql_instance_name,
            db_type=database_type,
            ip_address=get_ip_address_task,
            port=db_provider["port"],
        )

        @task(task_id=f"create_ssl_certificates_{database_type}")
        def create_ssl_certificate(instance: str, connection_id: str) -> dict[str, Any]:
            hook = CloudSQLHook(api_version="v1", gcp_conn_id=connection_id)
            certificate_name = f"test_cert_{''.join(random.choice(string.ascii_letters) for _ in range(8))}"
            response = hook.create_ssl_certificate(
                instance=instance,
                body={"common_name": certificate_name},
                project_id=PROJECT_ID,
            )
            return response

        create_ssl_certificate_task = create_ssl_certificate(
            instance=cloud_sql_instance_name, connection_id=create_connection_task
        )

        @task(task_id=f"save_ssl_cert_locally_{database_type}")
        def save_ssl_cert_locally(ssl_cert: dict[str, Any], db_type: str) -> dict[str, str]:
            folder = SSL_COMPOSER_PATH_PREFIX if run_in_composer() else SSL_LOCAL_PATH_PREFIX
            folder += f"/certs/{db_type}/{ssl_cert['operation']['name']}"
            if not os.path.exists(folder):
                os.makedirs(folder)
            _ssl_root_cert_path = f"{folder}/sslrootcert.pem"
            _ssl_cert_path = f"{folder}/sslcert.pem"
            _ssl_key_path = f"{folder}/sslkey.pem"
            with open(_ssl_root_cert_path, "w") as ssl_root_cert_file:
                ssl_root_cert_file.write(ssl_cert["serverCaCert"]["cert"])
            with open(_ssl_cert_path, "w") as ssl_cert_file:
                ssl_cert_file.write(ssl_cert["clientCert"]["certInfo"]["cert"])
            with open(_ssl_key_path, "w") as ssl_key_file:
                ssl_key_file.write(ssl_cert["clientCert"]["certPrivateKey"])
            return {
                "sslrootcert": _ssl_root_cert_path,
                "sslcert": _ssl_cert_path,
                "sslkey": _ssl_key_path,
            }

        save_ssl_cert_locally_task = save_ssl_cert_locally(
            ssl_cert=create_ssl_certificate_task, db_type=database_type
        )

        @task(task_id=f"save_ssl_cert_to_secret_manager_{database_type}")
        def save_ssl_cert_to_secret_manager(ssl_cert: dict[str, Any], db_type: str) -> str:
            hook = GoogleCloudSecretManagerHook()
            payload = {
                "sslrootcert": ssl_cert["serverCaCert"]["cert"],
                "sslcert": ssl_cert["clientCert"]["certInfo"]["cert"],
                "sslkey": ssl_cert["clientCert"]["certPrivateKey"],
            }
            _secret_id = f"secret_{DAG_ID}_{ENV_ID}_{db_type}"

            if not hook.secret_exists(project_id=PROJECT_ID, secret_id=_secret_id):
                hook.create_secret(
                    secret_id=_secret_id,
                    project_id=PROJECT_ID,
                )

            hook.add_secret_version(
                project_id=PROJECT_ID,
                secret_id=_secret_id,
                secret_payload=dict(data=base64.b64encode(json.dumps(payload).encode("ascii"))),
            )

            return _secret_id

        save_ssl_cert_to_secret_manager_task = save_ssl_cert_to_secret_manager(
            ssl_cert=create_ssl_certificate_task, db_type=database_type
        )

        task_id = f"example_cloud_sql_query_ssl_{database_type}"
        ssl_server_cert_path = (
            f"{{{{ task_instance.xcom_pull('save_ssl_cert_locally_{database_type}')['sslrootcert'] }}}}"
        )
        ssl_cert_path = (
            f"{{{{ task_instance.xcom_pull('save_ssl_cert_locally_{database_type}')['sslcert'] }}}}"
        )
        ssl_key_path = f"{{{{ task_instance.xcom_pull('save_ssl_cert_locally_{database_type}')['sslkey'] }}}}"

        # [START howto_operator_cloudsql_query_operators_ssl]
        query_task = CloudSQLExecuteQueryOperator(
            gcp_cloudsql_conn_id=conn_id,
            task_id=task_id,
            sql=SQL,
            ssl_client_cert=ssl_cert_path,
            ssl_server_cert=ssl_server_cert_path,
            ssl_client_key=ssl_key_path,
        )
        # [END howto_operator_cloudsql_query_operators_ssl]

        task_id = f"example_cloud_sql_query_ssl_secret_{database_type}"
        secret_id = f"{{{{ task_instance.xcom_pull('save_ssl_cert_to_secret_manager_{database_type}') }}}}"

        # [START howto_operator_cloudsql_query_operators_ssl_secret_id]
        query_task_secret = CloudSQLExecuteQueryOperator(
            gcp_cloudsql_conn_id=conn_id,
            task_id=task_id,
            sql=SQL,
            ssl_secret_id=secret_id,
        )
        # [END howto_operator_cloudsql_query_operators_ssl_secret_id]

        delete_instance = CloudSQLDeleteInstanceOperator(
            task_id=f"delete_cloud_sql_instance_{database_type}",
            project_id=PROJECT_ID,
            instance=cloud_sql_instance_name,
            trigger_rule=TriggerRule.ALL_DONE,
        )

        @task(task_id=f"delete_connection_{database_type}")
        def delete_connection(connection_id: str) -> None:
            delete_airflow_connection(connection_id=connection_id)

        delete_connection_task = delete_connection(connection_id=conn_id)

        @task(task_id=f"delete_secret_{database_type}")
        def delete_secret(ssl_secret_id, db_type: str) -> None:
            hook = GoogleCloudSecretManagerHook()
            if hook.secret_exists(project_id=PROJECT_ID, secret_id=ssl_secret_id):
                hook.delete_secret(secret_id=ssl_secret_id, project_id=PROJECT_ID)

        delete_secret_task = delete_secret(
            ssl_secret_id=save_ssl_cert_to_secret_manager_task, db_type=database_type
        )

        (
            # TEST SETUP
            create_cloud_sql_instance
            >> [create_database, create_user_task, get_ip_address_task]
            >> create_connection_task
            >> create_ssl_certificate_task
            >> [save_ssl_cert_locally_task, save_ssl_cert_to_secret_manager_task]
            # TEST BODY
            >> query_task
            >> query_task_secret
            # TEST TEARDOWN
            >> [delete_instance, delete_connection_task, delete_secret_task]
        )

    # ### Everything below this line is not part of example ###
    # ### Just for system tests purpose ###
    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""
This module contains various unit tests for Google Cloud DLP Operators
"""

from __future__ import annotations

from unittest import mock

from google.api_core.gapic_v1.method import DEFAULT
from google.cloud.dlp_v2.types import (
    DeidentifyContentResponse,
    DeidentifyTemplate,
    DlpJob,
    InspectContentResponse,
    InspectTemplate,
    JobTrigger,
    ListInfoTypesResponse,
    RedactImageResponse,
    ReidentifyContentResponse,
    StoredInfoType,
)

from airflow.providers.google.cloud.operators.dlp import (
    CloudDLPCancelDLPJobOperator,
    CloudDLPCreateDeidentifyTemplateOperator,
    CloudDLPCreateDLPJobOperator,
    CloudDLPCreateInspectTemplateOperator,
    CloudDLPCreateJobTriggerOperator,
    CloudDLPCreateStoredInfoTypeOperator,
    CloudDLPDeidentifyContentOperator,
    CloudDLPDeleteDeidentifyTemplateOperator,
    CloudDLPDeleteDLPJobOperator,
    CloudDLPDeleteInspectTemplateOperator,
    CloudDLPDeleteJobTriggerOperator,
    CloudDLPDeleteStoredInfoTypeOperator,
    CloudDLPGetDeidentifyTemplateOperator,
    CloudDLPGetDLPJobOperator,
    CloudDLPGetDLPJobTriggerOperator,
    CloudDLPGetInspectTemplateOperator,
    CloudDLPGetStoredInfoTypeOperator,
    CloudDLPInspectContentOperator,
    CloudDLPListDeidentifyTemplatesOperator,
    CloudDLPListDLPJobsOperator,
    CloudDLPListInfoTypesOperator,
    CloudDLPListInspectTemplatesOperator,
    CloudDLPListJobTriggersOperator,
    CloudDLPListStoredInfoTypesOperator,
    CloudDLPRedactImageOperator,
    CloudDLPReidentifyContentOperator,
    CloudDLPUpdateDeidentifyTemplateOperator,
    CloudDLPUpdateInspectTemplateOperator,
    CloudDLPUpdateJobTriggerOperator,
    CloudDLPUpdateStoredInfoTypeOperator,
)

GCP_CONN_ID = "google_cloud_default"
ORGANIZATION_ID = "test-org"
PROJECT_ID = "test-project"
DLP_JOB_ID = "job123"
TEMPLATE_ID = "template123"
STORED_INFO_TYPE_ID = "type123"
TRIGGER_ID = "trigger123"
DLP_JOB_PATH = f"projects/{PROJECT_ID}/dlpJobs/{DLP_JOB_ID}"
DLP_JOB_TRIGGER_PATH = f"projects/{PROJECT_ID}/jobTriggers/{TRIGGER_ID}"


class TestCloudDLPCancelDLPJobOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_cancel_dlp_job(self, mock_hook):
        mock_hook.return_value.cancel_dlp_job.return_value = mock.MagicMock()
        operator = CloudDLPCancelDLPJobOperator(dlp_job_id=DLP_JOB_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.cancel_dlp_job.assert_called_once_with(
            dlp_job_id=DLP_JOB_ID,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPCreateDeidentifyTemplateOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_create_deidentify_template(self, mock_hook):
        mock_hook.return_value.create_deidentify_template.return_value = mock.MagicMock()
        mock_hook.return_value.create_deidentify_template.return_value = DeidentifyTemplate(name=DLP_JOB_PATH)
        operator = CloudDLPCreateDeidentifyTemplateOperator(organization_id=ORGANIZATION_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.create_deidentify_template.assert_called_once_with(
            organization_id=ORGANIZATION_ID,
            project_id=None,
            deidentify_template=None,
            template_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPCreateDLPJobOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_create_dlp_job(self, mock_hook):
        mock_hook.return_value.create_dlp_job.return_value = DlpJob(
            name=DLP_JOB_PATH, state=DlpJob.JobState.PENDING
        )
        operator = CloudDLPCreateDLPJobOperator(project_id=PROJECT_ID, task_id="id")
        operator.execute(context=mock.MagicMock())

        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.create_dlp_job.assert_called_once_with(
            project_id=PROJECT_ID,
            inspect_job=None,
            risk_job=None,
            job_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
            wait_until_finished=True,
        )


class TestCloudDLPCreateInspectTemplateOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_create_inspect_template(self, mock_hook):
        mock_hook.return_value.create_inspect_template.return_value = InspectTemplate(name=DLP_JOB_PATH)
        operator = CloudDLPCreateInspectTemplateOperator(organization_id=ORGANIZATION_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.create_inspect_template.assert_called_once_with(
            organization_id=ORGANIZATION_ID,
            project_id=None,
            inspect_template=None,
            template_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPCreateJobTriggerOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_create_job_trigger(self, mock_hook):
        mock_hook.return_value.create_job_trigger.return_value = JobTrigger(name=DLP_JOB_TRIGGER_PATH)
        operator = CloudDLPCreateJobTriggerOperator(project_id=PROJECT_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.create_job_trigger.assert_called_once_with(
            project_id=PROJECT_ID,
            job_trigger=None,
            trigger_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPCreateStoredInfoTypeOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_create_stored_info_type(self, mock_hook):
        mock_hook.return_value.create_stored_info_type.return_value = StoredInfoType(name=DLP_JOB_PATH)
        operator = CloudDLPCreateStoredInfoTypeOperator(organization_id=ORGANIZATION_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.create_stored_info_type.assert_called_once_with(
            organization_id=ORGANIZATION_ID,
            project_id=None,
            config=None,
            stored_info_type_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPDeidentifyContentOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_deidentify_content(self, mock_hook):
        mock_hook.return_value.deidentify_content.return_value = DeidentifyContentResponse()
        operator = CloudDLPDeidentifyContentOperator(project_id=PROJECT_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.deidentify_content.assert_called_once_with(
            project_id=PROJECT_ID,
            deidentify_config=None,
            inspect_config=None,
            item=None,
            inspect_template_name=None,
            deidentify_template_name=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPDeleteDeidentifyTemplateOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_delete_deidentify_template(self, mock_hook):
        mock_hook.return_value.delete_deidentify_template.return_value = mock.MagicMock()
        operator = CloudDLPDeleteDeidentifyTemplateOperator(
            template_id=TEMPLATE_ID, organization_id=ORGANIZATION_ID, task_id="id"
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_deidentify_template.assert_called_once_with(
            template_id=TEMPLATE_ID,
            organization_id=ORGANIZATION_ID,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPDeleteDlpJobOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_delete_dlp_job(self, mock_hook):
        mock_hook.return_value.delete_dlp_job.return_value = mock.MagicMock()
        operator = CloudDLPDeleteDLPJobOperator(dlp_job_id=DLP_JOB_ID, project_id=PROJECT_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_dlp_job.assert_called_once_with(
            dlp_job_id=DLP_JOB_ID,
            project_id=PROJECT_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPDeleteInspectTemplateOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_delete_inspect_template(self, mock_hook):
        mock_hook.return_value.delete_inspect_template.return_value = mock.MagicMock()
        operator = CloudDLPDeleteInspectTemplateOperator(
            template_id=TEMPLATE_ID, organization_id=ORGANIZATION_ID, task_id="id"
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_inspect_template.assert_called_once_with(
            template_id=TEMPLATE_ID,
            organization_id=ORGANIZATION_ID,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPDeleteJobTriggerOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_delete_job_trigger(self, mock_hook):
        mock_hook.return_value.delete_job_trigger.return_value = mock.MagicMock()
        operator = CloudDLPDeleteJobTriggerOperator(
            job_trigger_id=TRIGGER_ID, project_id=PROJECT_ID, task_id="id"
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_job_trigger.assert_called_once_with(
            job_trigger_id=TRIGGER_ID,
            project_id=PROJECT_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPDeleteStoredInfoTypeOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_delete_stored_info_type(self, mock_hook):
        mock_hook.return_value.delete_stored_info_type.return_value = mock.MagicMock()
        operator = CloudDLPDeleteStoredInfoTypeOperator(
            stored_info_type_id=STORED_INFO_TYPE_ID,
            organization_id=ORGANIZATION_ID,
            task_id="id",
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.delete_stored_info_type.assert_called_once_with(
            stored_info_type_id=STORED_INFO_TYPE_ID,
            organization_id=ORGANIZATION_ID,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPGetDeidentifyTemplateOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_get_deidentify_template(self, mock_hook):
        mock_hook.return_value.get_deidentify_template.return_value = DeidentifyTemplate()
        operator = CloudDLPGetDeidentifyTemplateOperator(
            template_id=TEMPLATE_ID, organization_id=ORGANIZATION_ID, task_id="id"
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.get_deidentify_template.assert_called_once_with(
            template_id=TEMPLATE_ID,
            organization_id=ORGANIZATION_ID,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPGetDlpJobOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_get_dlp_job(self, mock_hook):
        mock_hook.return_value.get_dlp_job.return_value = DlpJob()
        operator = CloudDLPGetDLPJobOperator(dlp_job_id=DLP_JOB_ID, project_id=PROJECT_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.get_dlp_job.assert_called_once_with(
            dlp_job_id=DLP_JOB_ID,
            project_id=PROJECT_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPGetInspectTemplateOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_get_inspect_template(self, mock_hook):
        mock_hook.return_value.get_inspect_template.return_value = InspectTemplate()
        operator = CloudDLPGetInspectTemplateOperator(
            template_id=TEMPLATE_ID, organization_id=ORGANIZATION_ID, task_id="id"
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.get_inspect_template.assert_called_once_with(
            template_id=TEMPLATE_ID,
            organization_id=ORGANIZATION_ID,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPGetJobTripperOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_get_job_trigger(self, mock_hook):
        mock_hook.return_value.get_job_trigger.return_value = JobTrigger()
        operator = CloudDLPGetDLPJobTriggerOperator(
            job_trigger_id=TRIGGER_ID, project_id=PROJECT_ID, task_id="id"
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.get_job_trigger.assert_called_once_with(
            job_trigger_id=TRIGGER_ID,
            project_id=PROJECT_ID,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPGetStoredInfoTypeOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_get_stored_info_type(self, mock_hook):
        mock_hook.return_value.get_stored_info_type.return_value = StoredInfoType()
        operator = CloudDLPGetStoredInfoTypeOperator(
            stored_info_type_id=STORED_INFO_TYPE_ID,
            organization_id=ORGANIZATION_ID,
            task_id="id",
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.get_stored_info_type.assert_called_once_with(
            stored_info_type_id=STORED_INFO_TYPE_ID,
            organization_id=ORGANIZATION_ID,
            project_id=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPInspectContentOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_inspect_content(self, mock_hook):
        inspect_template_name = "inspect_template_name/name"
        mock_hook.return_value.inspect_content.return_value = InspectContentResponse()
        operator = CloudDLPInspectContentOperator(
            project_id=PROJECT_ID, task_id="id", inspect_template_name=inspect_template_name
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.inspect_content.assert_called_once_with(
            project_id=PROJECT_ID,
            inspect_config=None,
            item=None,
            inspect_template_name=inspect_template_name,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPListDeidentifyTemplatesOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_list_deidentify_templates(self, mock_hook):
        mock_hook.return_value.list_deidentify_templates.return_value = mock.MagicMock()
        operator = CloudDLPListDeidentifyTemplatesOperator(organization_id=ORGANIZATION_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.list_deidentify_templates.assert_called_once_with(
            organization_id=ORGANIZATION_ID,
            project_id=None,
            page_size=None,
            order_by=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPListDlpJobsOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_list_dlp_jobs(self, mock_hook):
        mock_hook.return_value.list_dlp_jobs.return_value = mock.MagicMock()
        operator = CloudDLPListDLPJobsOperator(project_id=PROJECT_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.list_dlp_jobs.assert_called_once_with(
            project_id=PROJECT_ID,
            results_filter=None,
            page_size=None,
            job_type=None,
            order_by=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPListInfoTypesOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_list_info_types(self, mock_hook):
        mock_hook.return_value.list_info_types.return_value = ListInfoTypesResponse()
        operator = CloudDLPListInfoTypesOperator(task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.list_info_types.assert_called_once_with(
            language_code=None,
            results_filter=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPListInspectTemplatesOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_list_inspect_templates(self, mock_hook):
        mock_hook.return_value.list_inspect_templates.return_value = mock.MagicMock()
        operator = CloudDLPListInspectTemplatesOperator(organization_id=ORGANIZATION_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.list_inspect_templates.assert_called_once_with(
            organization_id=ORGANIZATION_ID,
            project_id=None,
            page_size=None,
            order_by=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPListJobTriggersOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_list_job_triggers(self, mock_hook):
        mock_hook.return_value.list_job_triggers.return_value = mock.MagicMock()
        operator = CloudDLPListJobTriggersOperator(project_id=PROJECT_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.list_job_triggers.assert_called_once_with(
            project_id=PROJECT_ID,
            page_size=None,
            order_by=None,
            results_filter=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPListStoredInfoTypesOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_list_stored_info_types(self, mock_hook):
        mock_hook.return_value.list_stored_info_types.return_value = mock.MagicMock()
        operator = CloudDLPListStoredInfoTypesOperator(organization_id=ORGANIZATION_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.list_stored_info_types.assert_called_once_with(
            organization_id=ORGANIZATION_ID,
            project_id=None,
            page_size=None,
            order_by=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPRedactImageOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_redact_image(self, mock_hook):
        mock_hook.return_value.redact_image.return_value = RedactImageResponse()
        operator = CloudDLPRedactImageOperator(project_id=PROJECT_ID, task_id="id")
        operator.execute(context=None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.redact_image.assert_called_once_with(
            project_id=PROJECT_ID,
            inspect_config=None,
            image_redaction_configs=None,
            include_findings=None,
            byte_item=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPReidentifyContentOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_reidentify_content(self, mock_hook):
        mock_hook.return_value.reidentify_content.return_value = ReidentifyContentResponse()
        operator = CloudDLPReidentifyContentOperator(project_id=PROJECT_ID, task_id="id")
        operator.execute(context=None)
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.reidentify_content.assert_called_once_with(
            project_id=PROJECT_ID,
            reidentify_config=None,
            inspect_config=None,
            item=None,
            inspect_template_name=None,
            reidentify_template_name=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPUpdateDeidentifyTemplateOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_update_deidentify_template(self, mock_hook):
        mock_hook.return_value.update_deidentify_template.return_value = DeidentifyTemplate()
        operator = CloudDLPUpdateDeidentifyTemplateOperator(
            template_id=TEMPLATE_ID, organization_id=ORGANIZATION_ID, task_id="id"
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.update_deidentify_template.assert_called_once_with(
            template_id=TEMPLATE_ID,
            organization_id=ORGANIZATION_ID,
            project_id=None,
            deidentify_template=None,
            update_mask=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPUpdateInspectTemplateOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_update_inspect_template(self, mock_hook):
        mock_hook.return_value.update_inspect_template.return_value = InspectTemplate()
        operator = CloudDLPUpdateInspectTemplateOperator(
            template_id=TEMPLATE_ID, organization_id=ORGANIZATION_ID, task_id="id"
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.update_inspect_template.assert_called_once_with(
            template_id=TEMPLATE_ID,
            organization_id=ORGANIZATION_ID,
            project_id=None,
            inspect_template=None,
            update_mask=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPUpdateJobTriggerOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_update_job_trigger(self, mock_hook):
        mock_hook.return_value.update_job_trigger.return_value = JobTrigger()
        operator = CloudDLPUpdateJobTriggerOperator(job_trigger_id=TRIGGER_ID, task_id="id")
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.update_job_trigger.assert_called_once_with(
            job_trigger_id=TRIGGER_ID,
            project_id=None,
            job_trigger=None,
            update_mask=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )


class TestCloudDLPUpdateStoredInfoTypeOperator:
    @mock.patch("airflow.providers.google.cloud.operators.dlp.CloudDLPHook")
    def test_update_stored_info_type(self, mock_hook):
        mock_hook.return_value.update_stored_info_type.return_value = StoredInfoType()
        operator = CloudDLPUpdateStoredInfoTypeOperator(
            stored_info_type_id=STORED_INFO_TYPE_ID,
            organization_id=ORGANIZATION_ID,
            task_id="id",
        )
        operator.execute(context=mock.MagicMock())
        mock_hook.assert_called_once_with(
            gcp_conn_id=GCP_CONN_ID,
            impersonation_chain=None,
        )
        mock_hook.return_value.update_stored_info_type.assert_called_once_with(
            stored_info_type_id=STORED_INFO_TYPE_ID,
            organization_id=ORGANIZATION_ID,
            project_id=None,
            config=None,
            update_mask=None,
            retry=DEFAULT,
            timeout=None,
            metadata=(),
        )
